"""stepsrunner.py unit tests."""
import logging
import pytest
from unittest.mock import call, patch
from pypyr.context import Context
from pypyr.dsl import Step
from pypyr.errors import (Call,
                          ContextError,
                          Jump,
                          Stop,
                          StopPipeline,
                          StopStepGroup)
from pypyr.pipeline import Pipeline
from pypyr.stepsrunner import StepsRunner
from tests.common.utils import DeepCopyMagicMock, patch_logger


# region test context


def arb_step_mock(context):
    """No real reason, other than to mock the existence of a run_step."""
    return 'from arb step mock'


def get_test_context():
    """Return a pypyr context for testing."""
    return Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'key4': [
            {'k4lk1': 'value4',
             'k4lk2': 'value5'},
            {'k4lk1': 'value6',
             'k4lk2': 'value7'}
        ],
        'key5': False,
        'key6': True,
        'key7': 77
    })


def get_test_pipeline():
    """Return an arbitrary pipeline definition."""
    return {
        'sg1': [
            'step1',
            'step2',
            {'step3key1': 'values3k1', 'step3key2': 'values3k2'},
            'step4'
        ],
        'sg2': False,
        'sg3': 77,
        'sg4': None
    }


def get_valid_test_pipeline():
    """Return an arbitrary pipeline definition."""
    return {
        'sg1': [
            'step1',
            'step2',
            {'name': 'step3key1',
             'in':
                {'in3k1_1': 'v3k1', 'in3k1_2': 'v3k2'}},
            'step4'
        ],
        'sg2': False,
        'sg3': 77,
        'sg4': None
    }


def get_valid_steps_pipeline():
    """Pipeline definition where steps point to actual test step."""
    return {
        'sg1': [
            'tests.arbpack.arbincrementstep',
            'tests.arbpack.arbincrementstep'],
        'sg2': [
            'tests.arbpack.arbincrementstep',
        ],
        'sg3': None
    }

# endregion test context

# region init


def test_stepsrunner_init():
    """The StepsRunner initializes."""
    s = StepsRunner(pipeline_body=3, context=4)
    assert s.pipeline_body == 3
    assert s.context == 4

# endregion init

# region get_pipeline_steps


def test_get_pipeline_steps_pass():
    """Return named step group from pipeline."""
    with patch_logger('pypyr.stepsrunner', logging.DEBUG) as mock_logger_debug:
        steps = StepsRunner(get_test_pipeline(), {}).get_pipeline_steps('sg1')

    assert len(steps) == 4
    assert steps[0] == 'step1'
    assert steps[1] == 'step2'
    assert steps[2] == {'step3key1': 'values3k1', 'step3key2': 'values3k2'}
    assert steps[3] == 'step4'

    mock_logger_debug.assert_any_call("4 steps found under sg1 in "
                                      "pipeline definition.")


def test_get_pipeline_steps_not_found():
    """Can't find step group in pipeline."""
    with patch_logger('pypyr.stepsrunner', logging.DEBUG) as mock_logger_debug:
        steps = StepsRunner(get_test_pipeline(), {}).get_pipeline_steps('arb')
        assert steps is None

    mock_logger_debug.assert_any_call(
        "pipeline doesn't have a arb collection. Add a arb: sequence to the "
        "yaml if you want arb actually to do something.")


def test_get_pipeline_steps_none():
    """Find step group in pipeline but it has no steps."""
    with patch_logger(
            'pypyr.stepsrunner', logging.WARNING) as mock_logger_warn:
        steps = StepsRunner(get_test_pipeline(), {}).get_pipeline_steps('sg4')
        assert steps is None

    mock_logger_warn.assert_called_once_with(
        "sg4: sequence has no elements. So it won't do anything.")
# endregion get_pipeline_steps

# region run_failure_step_group


def test_run_failure_step_group_pass():
    """Failure step group runner passes on_failure to step group runner."""
    with patch.object(StepsRunner,
                      'run_step_group') as mock_run_group:
        StepsRunner({'pipe': 'val'},
                    Context()).run_failure_step_group('on_failure')

    mock_run_group.assert_called_once_with('on_failure', raise_stop=True)


def test_run_failure_step_group_swallows():
    """Failure step group runner swallows errors."""
    with patch.object(StepsRunner,
                      'run_step_group') as mock_run_group:
        with patch_logger(
                'pypyr.stepsrunner', logging.ERROR) as mock_logger_error:
            mock_run_group.side_effect = ContextError('arb error')
            runner = StepsRunner({'pipe': 'val'}, Context())
            runner.run_failure_step_group('arb_failure')

        mock_logger_error.assert_any_call(
            "Failure handler also failed. Swallowing.")

    assert runner.pipeline_body == {'pipe': 'val'}
    assert runner.context == Context()
    mock_run_group.assert_called_once_with('arb_failure', raise_stop=True)


def test_run_failure_step_group_stop():
    """Failure step group runner does Stop."""
    with patch.object(StepsRunner,
                      'run_step_group') as mock_run_group:
        with patch_logger(
                'pypyr.stepsrunner', logging.DEBUG) as mock_logger_debug:
            mock_run_group.side_effect = StopStepGroup
            runner = StepsRunner({'pipe': 'val'}, Context())

            with pytest.raises(StopStepGroup):
                runner.run_failure_step_group('arb_failure')

        mock_logger_debug.assert_any_call(
            "Stop instruction: done with failure handler arb_failure.")

    assert runner.pipeline_body == {'pipe': 'val'}
    assert runner.context == Context()
    mock_run_group.assert_called_once_with('arb_failure', raise_stop=True)

# endregion run_failure_step_group

# region run_pipeline_steps


def test_run_pipeline_steps_none():
    """If steps None does nothing."""
    with patch_logger('pypyr.stepsrunner', logging.DEBUG) as mock_logger_debug:
        StepsRunner(None, Context({'k1': 'v1'})).run_pipeline_steps(None)

    mock_logger_debug.assert_any_call("No steps found to execute.")


@patch('pypyr.moduleloader.get_module')
@patch.object(Step, 'invoke_step')
def test_run_pipeline_steps_complex(mock_invoke_step, mock_module):
    """Complex step run with no in args."""
    with patch_logger('pypyr.dsl', logging.DEBUG) as mock_logger_debug:
        StepsRunner(None, Context({'k1': 'v1'})).run_pipeline_steps(
            [{'name': 'step1'}])

    mock_logger_debug.assert_any_call("step1 is complex.")
    mock_invoke_step.assert_called_once_with(context={'k1': 'v1'})


@patch('pypyr.moduleloader.get_module')
@patch.object(Step, 'invoke_step')
@patch('unittest.mock.MagicMock', new=DeepCopyMagicMock)
def test_run_pipeline_steps_complex_with_in(mock_invoke_step, mock_module):
    """Complex step run with in args. In args added to context for run_step."""
    steps = [{
        'name': 'step1',
        'in': {'newkey1': 'v1',
               'newkey2': 'v2',
               'key3': 'updated in',
               'key4': [0, 1, 2, 3],
               'key5': True,
               'key6': False,
               'key7': 88}
    }]
    context = Context({
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3',
        'key4': [
            {'k4lk1': 'value4',
             'k4lk2': 'value5'},
            {'k4lk1': 'value6',
             'k4lk2': 'value7'}
        ],
        'key5': False,
        'key6': True,
        'key7': 77,
        'key8': None,
        'key9': 'arb'
    })

    with patch_logger('pypyr.stepsrunner', logging.DEBUG) as mock_logger_debug:
        StepsRunner(None, context).run_pipeline_steps(steps)

    mock_logger_debug.assert_any_call("executed 1 steps")
    assert mock_invoke_step.call_count == 1
    # step invoked with the context updated from 'in'
    assert mock_invoke_step.call_args_list[0] == call(
        context={'key1': 'value1',
                 'key2': 'value2',
                 'key3': 'updated in',
                 'key4': [0, 1, 2, 3],
                 'key5': True,
                 'key6': False,
                 'key7': 88,
                 'key8': None,
                 'key9': 'arb',
                 'newkey1': 'v1',
                 'newkey2': 'v2'})

    # all items keys 'in' purged from context.
    assert context == {'key1': 'value1',
                       'key2': 'value2',
                       'key8': None,
                       'key9': 'arb'}


@patch('pypyr.moduleloader.get_module')
@patch.object(Step, 'run_step')
def test_run_pipeline_steps_simple(mock_run_step, mock_module):
    """Simple step run."""
    with patch_logger('pypyr.dsl', logging.DEBUG) as mock_logger_debug:
        StepsRunner(None, {'k1': 'v1'}).run_pipeline_steps(['step1'])

    mock_logger_debug.assert_any_call('step1 is a simple string.')
    mock_run_step.assert_called_once_with({'k1': 'v1'})


# region run_pipeline_steps: run


@patch('pypyr.moduleloader.get_module')
@patch.object(Step, 'invoke_step')
def test_run_pipeline_steps_mix_run_and_not_run(mock_invoke_step, mock_module):
    """Complex steps, some run some don't."""
    # Step 1 & 3 runs, 2 should not.
    steps = [{
        'name': 'step1',
        'run': True
    },
        {
        'name': 'step2',
        'run': False
    },
        {
        'name': 'step3',
    },
    ]

    context = get_test_context()
    original_len = len(context)

    with patch_logger('pypyr.stepsrunner', logging.DEBUG) as mock_logger_debug:
        with patch_logger('pypyr.dsl', logging.INFO) as mock_logger_info:
            StepsRunner(None, context).run_pipeline_steps(steps)

    mock_logger_debug.assert_any_call("executed 3 steps")
    mock_logger_info.assert_any_call(
        "step2 not running because run is False.")

    assert mock_invoke_step.call_count == 2
    assert mock_invoke_step.mock_calls == [call(context={
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3', 'key4':
        [
            {
                'k4lk1': 'value4',
                'k4lk2': 'value5'},
            {
                'k4lk1': 'value6',
                'k4lk2': 'value7'
            }],
        'key5': False,
        'key6': True,
        'key7': 77}),
        call(context={
            'key1': 'value1',
            'key2': 'value2',
            'key3': 'value3', 'key4':
            [
                {
                    'k4lk1': 'value4',
                    'k4lk2': 'value5'},
                {
                    'k4lk1': 'value6',
                    'k4lk2': 'value7'
                }],
            'key5': False,
            'key6': True,
            'key7': 77})]

    # validate all the in params ended up in context as intended
    assert len(context) == original_len


@patch('pypyr.moduleloader.get_module')
@patch.object(Step, 'invoke_step')
def test_run_pipeline_steps_complex_with_multistep_none_run(mock_invoke_step,
                                                            mock_module):
    """Multiple steps and none run."""
    # None of these should run - various shades of python false
    steps = [{
        'name': 'step1',
        'run': False
    },
        {
        'name': 'step2',
        'run': 0
    },
        {
        'name': 'step3',
        'run': None,
    },
    ]

    context = get_test_context()
    original_len = len(context)

    with patch_logger('pypyr.dsl', logging.INFO) as mock_logger_info:
        StepsRunner(None, context).run_pipeline_steps(steps)

    mock_logger_info.assert_any_call(
        "step1 not running because run is False.")
    mock_logger_info.assert_any_call(
        "step2 not running because run is False.")
    mock_logger_info.assert_any_call(
        "step3 not running because run is False.")
    mock_invoke_step.assert_not_called()

    # validate all the in params ended up in context as intended
    assert len(context) == original_len
# endregion  run_pipeline_steps: run

# region  run_pipeline_steps: skip


@patch('pypyr.moduleloader.get_module')
@patch.object(Step, 'invoke_step')
def test_run_pipeline_steps_mix_skip_and_not_skip(mock_invoke_step,
                                                  mock_module):
    """Complex steps, some run some don't."""
    # Step 1 & 3 runs, 2 should not.
    steps = [{
        'name': 'step1',
        'skip': False
    },
        {
        'name': 'step2',
        'skip': True
    },
        {
        'name': 'step3',
    },
    ]

    context = get_test_context()
    original_len = len(context)

    with patch_logger('pypyr.stepsrunner', logging.DEBUG) as mock_logger_debug:
        with patch_logger('pypyr.dsl', logging.INFO) as mock_logger_info:
            StepsRunner(None, context).run_pipeline_steps(steps)

    mock_logger_debug.assert_any_call("executed 3 steps")
    mock_logger_info.assert_any_call(
        "step2 not running because skip is True.")

    assert mock_invoke_step.call_count == 2
    assert mock_invoke_step.mock_calls == [call(context={
        'key1': 'value1',
        'key2': 'value2',
        'key3': 'value3', 'key4':
        [
            {
                'k4lk1': 'value4',
                'k4lk2': 'value5'},
            {
                'k4lk1': 'value6',
                'k4lk2': 'value7'
            }],
        'key5': False,
        'key6': True,
        'key7': 77}),
        call(context={
            'key1': 'value1',
            'key2': 'value2',
            'key3': 'value3', 'key4':
            [
                {
                    'k4lk1': 'value4',
                    'k4lk2': 'value5'},
                {
                    'k4lk1': 'value6',
                    'k4lk2': 'value7'
                }],
            'key5': False,
            'key6': True,
            'key7': 77})]

    # validate all the in params ended up in context as intended
    assert len(context) == original_len


@patch('pypyr.moduleloader.get_module')
@patch.object(Step, 'invoke_step')
def test_run_pipeline_steps_complex_with_multistep_all_skip(mock_invoke_step,
                                                            mock_module):
    """Multiple steps and none run."""
    # None of these should run - various shades of python true
    steps = [{
        'name': 'step1',
        'skip': [1, 2, 3]
    },
        {
        'name': 'step2',
        'skip': 1
    },
        {
        'name': 'step3',
        'skip': True,
    },
        {
        'name': 'step4',
        # run evals before skip
        'run': False,
        'skip': False
    },
    ]

    context = get_test_context()
    original_len = len(context)

    with patch_logger('pypyr.dsl', logging.INFO) as mock_logger_info:
        StepsRunner(None, context).run_pipeline_steps(steps)

    mock_logger_info.assert_any_call(
        "step1 not running because skip is True.")
    mock_logger_info.assert_any_call(
        "step2 not running because skip is True.")
    mock_logger_info.assert_any_call(
        "step3 not running because skip is True.")
    mock_logger_info.assert_any_call(
        "step4 not running because run is False.")
    mock_invoke_step.assert_not_called()

    # validate all the in params ended up in context as intended
    assert len(context) == original_len

# endregion run_pipeline_steps: skip

# region run_pipeline_steps: swallow


@patch('pypyr.moduleloader.get_module')
@patch.object(Step, 'invoke_step')
@patch('unittest.mock.MagicMock', new=DeepCopyMagicMock)
def test_run_pipeline_steps_swallow_sequence(mock_invoke_step, mock_module):
    """Complex steps, some run some don't, some swallow, some don't."""
    step4_run_error_swallow = {
        'col': None,
        'customError': {},
        'description': 'arb error here 4',
        'exception': ValueError('arb error here 4'),
        'line': None,
        'name': 'ValueError',
        'step': 'step4',
        'swallowed': True,
    }

    step6_run_error_raise = {
        'col': None,
        'customError': {},
        'description': 'arb error here 6',
        'exception': ValueError('arb error here 6'),
        'line': None,
        'name': 'ValueError',
        'step': 'step6',
        'swallowed': False,
    }

    # 1 & 2 don't run,
    # 3 runs, no error
    # 4 runs, error and swallows
    # 5 skips
    # 6 runs, error and raises
    # 7 doesn't run because execution stopped at 6 because of error
    mock_invoke_step.side_effect = [
        None,  # 3
        step4_run_error_swallow['exception'],  # 4
        step6_run_error_raise['exception'],  # 6
    ]

    steps = [
        {
            'name': 'step1',
            'run': False
        },
        {
            'name': 'step2',
            'skip': True
        },
        {
            'name': 'step3',
        },
        {
            'name': 'step4',
            'run': True,
            'skip': False,
            'swallow': 1
        },
        {
            'name': 'step5',
            'run': True,
            'skip': True,
            'swallow': True
        },
        'step6',
        'step7',
    ]

    context = get_test_context()
    original_len = len(context)

    with patch_logger('pypyr.dsl', logging.DEBUG) as mock_logger_debug:
        with patch_logger('pypyr.dsl', logging.INFO) as mock_logger_info:
            with patch_logger('pypyr.dsl', logging.ERROR) as mock_logger_error:
                with pytest.raises(ValueError) as err_info:
                    StepsRunner(None, context).run_pipeline_steps(steps)

                    assert str(err_info.value) == "arb error here 6"

    mock_logger_debug.assert_has_calls([
        call('step1 is complex.'),
        call('step2 is complex.'),
        call('step3 is complex.'),
        call('step4 is complex.'),
        call('step5 is complex.'),
        call('step6 is a simple string.')],
        any_order=True
    )

    assert mock_logger_info.call_args_list == [
        call('step1 not running because run is False.'),
        call('step2 not running because skip is True.'),
        call('step5 not running because skip is True.')]

    assert mock_invoke_step.call_count == 3

    assert mock_logger_error.call_args_list == [
        call(
            "step4 Ignoring error because swallow is True "
            "for this step.\n"
            "ValueError: arb error here 4"
        ),
        call("Error while running step step6"),
    ]

    assert mock_invoke_step.call_count == 3

    # step3
    assert mock_invoke_step.call_args_list[0] == call(
        context=get_test_context()
    )

    # step4
    assert mock_invoke_step.call_args_list[1] == call(
        context=get_test_context()
    )

    step6_context = mock_invoke_step.call_args_list[2][1]["context"]
    assert get_test_context().items() <= step6_context.items()
    assert len(step6_context['runErrors']) == 1

    # validate all the in params ended up in context as intended,
    # plus runErrors
    assert len(context) == original_len + 1
    assert len(context['runErrors']) == 2

    assert step4_run_error_swallow == context['runErrors'][0]
    assert step6_run_error_raise == context['runErrors'][1]

# endregion run_pipeline_steps: swallow


# endregion run_pipeline_steps

# region run_step_group

@patch.object(StepsRunner, 'run_pipeline_steps')
def test_run_step_group_pass(mock_run_steps):
    """run_step_group gets and runs steps for group."""
    StepsRunner(get_valid_test_pipeline(), Context()).run_step_group(
        step_group_name='sg1')

    mock_run_steps.assert_called_once_with(steps=[
        'step1',
        'step2',
        {'name': 'step3key1',
         'in':
            {'in3k1_1': 'v3k1', 'in3k1_2': 'v3k2'}},
        'step4'
    ])


@patch.object(StepsRunner, 'run_pipeline_steps')
def test_run_step_group_no_raise(mock_run_steps):
    """run_step_group with raise_stop False."""
    mock_run_steps.side_effect = StopStepGroup()
    StepsRunner(get_valid_test_pipeline(),
                Context()).run_step_group(step_group_name='sg1',
                                          raise_stop=False)

    mock_run_steps.assert_called_once_with(steps=[
        'step1',
        'step2',
        {'name': 'step3key1',
         'in':
            {'in3k1_1': 'v3k1', 'in3k1_2': 'v3k2'}},
        'step4'
    ])


@patch.object(StepsRunner, 'run_pipeline_steps')
def test_run_step_group_raise(mock_run_steps):
    """run_step_group with raise_stop True."""
    mock_run_steps.side_effect = StopStepGroup()
    with pytest.raises(StopStepGroup):
        StepsRunner(get_valid_test_pipeline(),
                    Context()).run_step_group(step_group_name='sg1',
                                              raise_stop=True)

    mock_run_steps.assert_called_once_with(steps=[
        'step1',
        'step2',
        {'name': 'step3key1',
         'in':
            {'in3k1_1': 'v3k1', 'in3k1_2': 'v3k2'}},
        'step4'
    ])

# endregion run_step_group

# region run_step_groups


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_single(mock_run_step_group):
    """Single test group runs with success."""
    StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
        groups=['sg1'],
        success_group='arb success',
        failure_group='arb fail')

    assert mock_run_step_group.mock_calls == [call('sg1'),
                                              call('arb success')]


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_single_no_success_handler(mock_run_step_group):
    """Single test group runs with no success handler."""
    StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
        groups=['sg1'],
        success_group=None,
        failure_group='arb fail')

    assert mock_run_step_group.mock_calls == [call('sg1')]


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_sequence(mock_run_step_group):
    """Sequence of test groups runs with success."""
    StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
        groups=['sg3', 'sg1', 'sg2', 'sg4'],
        success_group='arb success',
        failure_group='arb fail')

    assert mock_run_step_group.mock_calls == [call('sg3'),
                                              call('sg1'),
                                              call('sg2'),
                                              call('sg4'),
                                              call('arb success')]


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_sequence_with_fail(mock_run_step_group):
    """Sequence of test groups runs with failure."""
    mock_run_step_group.side_effect = [None, None, ValueError('arb')]

    with pytest.raises(ValueError) as err:
        StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
            groups=['sg3', 'sg1', 'sg2', 'sg4'],
            success_group='arb success',
            failure_group='arb fail')

    assert str(err.value) == 'arb'
    assert mock_run_step_group.mock_calls == [call('sg3'),
                                              call('sg1'),
                                              call('sg2'),
                                              call('arb fail', raise_stop=True)
                                              ]


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_sequence_with_failing_fail(mock_run_step_group):
    """Sequence of test groups runs with failure, failure handler fails."""
    mock_run_step_group.side_effect = [None,
                                       None,
                                       ValueError('arb'),
                                       KeyError('arb failure handler err')]

    with pytest.raises(ValueError) as err:
        StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
            groups=['sg3', 'sg1', 'sg2', 'sg4'],
            success_group='arb success',
            failure_group='arb fail')

    # failure handler swallows arb KeyError
    assert str(err.value) == 'arb'
    assert mock_run_step_group.mock_calls == [call('sg3'),
                                              call('sg1'),
                                              call('sg2'),
                                              call('arb fail', raise_stop=True)
                                              ]


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_sequence_with_failing_stop(
        mock_run_step_group):
    """Groups run with failure, failure handler Stop."""
    mock_run_step_group.side_effect = [None,
                                       None,
                                       ValueError('arb'),
                                       Stop]

    with pytest.raises(Stop):
        StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
            groups=['sg3', 'sg1', 'sg2', 'sg4'],
            success_group='arb success',
            failure_group='arb fail')

    # failure handler swallows arb KeyError
    assert mock_run_step_group.mock_calls == [call('sg3'),
                                              call('sg1'),
                                              call('sg2'),
                                              call('arb fail', raise_stop=True)
                                              ]


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_sequence_with_failing_stoppipeline(
        mock_run_step_group):
    """Groups run with failure, failure handler StopPipeline."""
    mock_run_step_group.side_effect = [None,
                                       None,
                                       ValueError('arb'),
                                       StopPipeline]

    with pytest.raises(StopPipeline):
        StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
            groups=['sg3', 'sg1', 'sg2', 'sg4'],
            success_group='arb success',
            failure_group='arb fail')

    # failure handler swallows arb KeyError
    assert mock_run_step_group.mock_calls == [call('sg3'),
                                              call('sg1'),
                                              call('sg2'),
                                              call('arb fail', raise_stop=True)
                                              ]


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_sequence_with_failing_stopstepgroup(
        mock_run_step_group):
    """Test groups runs with failure, failure handler raises StopStepGroup."""
    mock_run_step_group.side_effect = [None,
                                       None,
                                       ValueError('arb'),
                                       StopStepGroup]

    StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
        groups=['sg3', 'sg1', 'sg2', 'sg4'],
        success_group='arb success',
        failure_group='arb fail')

    # failure handler swallows arb ValueError
    assert mock_run_step_group.mock_calls == [call('sg3'),
                                              call('sg1'),
                                              call('sg2'),
                                              call('arb fail', raise_stop=True)
                                              ]


@patch.object(StepsRunner, 'run_step_group')
def test_run_step_groups_sequence_with_fail_no_handler(mock_run_step_group):
    """Sequence of test groups runs with failure when no failure handler."""
    mock_run_step_group.side_effect = [None, None, ValueError('arb')]

    with pytest.raises(ValueError) as err:
        StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
            groups=['sg3', 'sg1', 'sg2', 'sg4'],
            success_group='arb success',
            failure_group=None)

    assert str(err.value) == 'arb'
    assert mock_run_step_group.mock_calls == [call('sg3'),
                                              call('sg1'),
                                              call('sg2')]


def test_run_step_groups_sequence_with_mutate():
    """Sequence of test groups runs with success and mutates context."""
    context = Context({'counter': 5})
    StepsRunner(get_valid_steps_pipeline(), context).run_step_groups(
        groups=['sg3', 'sg1', 'sg2', 'sg4'],
        success_group='arb success',
        failure_group='arb fail')

    assert context['counter'] == 8


def test_run_step_groups_none_groups():
    """Raise error if groups None."""
    with pytest.raises(ValueError) as err:
        StepsRunner(get_valid_test_pipeline(), Context()).run_step_groups(
            groups=None,
            success_group='arb success',
            failure_group='arb fail')

    assert str(err.value) == (
        'you must specify which step-groups you want to run. groups is None.')
# endregion run_step_groups

# region Jump


def get_jump_pipeline():
    """Test pipeline for jump."""
    return {
        'sg1': [
            'sg1.step1',
            'sg1.step2'
        ],
        'sg2': [
            'sg2.step1',
            'sg2.step2'
        ],
        'sg3': [
            'sg3.step1',
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


def nothing_step(context):
    """Do nothing."""
    pass


def jump_step(groups, success=None, failure=None):
    """Jump step mock."""
    def run_step(context):
        raise Jump(groups, success, failure, 'jumparb')
    return run_step


def call_step(groups,
              success=None,
              failure=None,
              original_config=('call', 'arb')):
    """Call step mock."""
    def run_step(context):
        raise Call(groups, success, failure, original_config)
    return run_step


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_jump_with_success_handler(mock_step_cache):
    """Jump between different step groups with success handler."""
    # Sequence: sg2 - sg2.1 (JUMP)
    #           sg1 - sg1.1, sg 1.2 (JUMP)
    #           sg4 - sg4.1 (JUMP)
    #           sg3 - sg3.1, sg 3.2
    #           sg5 - sg5.1 (on_success)
    mock_step_cache.side_effect = [
        jump_step(['sg1']),  # 2.1
        nothing_step,  # 1.1
        jump_step(['sg4']),  # 1.2
        jump_step(['sg3']),  # 4.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
        nothing_step,  # 5.1
    ]

    context = Context()
    StepsRunner(get_jump_pipeline(), context).run_step_groups(
        groups=['sg2'],
        success_group='sg5',
        failure_group=None)

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg1.step1'),
                                          call('sg1.step2'),
                                          call('sg4.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2'),
                                          call('sg5.step1')]


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_jump_with_failure_handler(mock_step_cache):
    """Jump between different step groups with failure handler."""
    # Sequence: sg2 - sg2.1 (JUMP)
    #           sg3 - sg3.1 (ERROR)
    #           sg4 - sg4.1, sg4.2 (failure handler)
    def err_step(context):
        raise ValueError('3.1')

    mock_step_cache.side_effect = [
        jump_step(['sg3']),  # 2.1
        err_step,  # 3.1
        nothing_step,  # 4.1
        nothing_step,  # 4.2
    ]

    context = Context()
    with pytest.raises(ValueError) as err_info:
        StepsRunner(get_jump_pipeline(), context).run_step_groups(
            groups=['sg2', 'sg1'],
            success_group='sg5',
            failure_group='sg4')

    assert str(err_info.value) == '3.1'
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg3.step1'),
                                          call('sg4.step1'),
                                          call('sg4.step2')]


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_jump_with_group_sequences_and_success_jump(mock_step_cache):
    """Jump between step groups with success handler and jump in success."""
    # Sequence: sg2 - sg2.1, sg2.2
    #           sg1 - sg1.1, sg 1.2 (JUMP)
    #           sg4 - sg4.1, sg 4.2
    #           sg3 - sg3.1, sg 3.2
    #           sg6 - sg6.1 (JUMP) - on_success
    #           sg5 - sg5.1
    mock_step_cache.side_effect = [
        nothing_step,  # 2.1
        nothing_step,  # 2.2
        nothing_step,  # 1.1
        jump_step(['sg4', 'sg3']),  # 1.2
        nothing_step,  # 4.1
        nothing_step,  # 4.2
        nothing_step,  # 3.1
        nothing_step,  # 3.2
        jump_step(['sg5']),  # 6.1
        nothing_step,  # 5.1
        nothing_step,  # 5.2
    ]

    context = Context()
    StepsRunner(get_jump_pipeline(), context).run_step_groups(
        groups=['sg2', 'sg1'],
        success_group='sg6',
        failure_group=None)

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg1.step1'),
                                          call('sg1.step2'),
                                          call('sg4.step1'),
                                          call('sg4.step2'),
                                          call('sg3.step1'),
                                          call('sg3.step2'),
                                          call('sg6.step1'),
                                          call('sg5.step1')]


def get_while_pipeline():
    """Test pipeline for while."""
    return {
        'sg1': [
            'sg1.step1',
            'sg1.step2'
        ],
        'sg2': [
            {'name': 'sg2.step1',
             'while': {
                 'max': 3}
             },
            'sg2.step2'
        ],
        'sg3': [
            'sg3.step1',
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_jump_with_while(mock_step_cache):
    """Jump between different step groups in a while."""
    # Sequence: sg2 - sg2.1 x2 (JUMP)
    #           sg1 - sg1.1, sg 1.2 (JUMP)
    #           sg4 - sg4.1 (JUMP)
    #           sg3 - sg3.1, sg 3.2
    #           sg5 - sg5.1 (on_success)

    mock21 = DeepCopyMagicMock()

    def step21(context):
        mock21(context)
        if context['whileCounter'] == 2:
            jump_step(['sg1'])(context)

    mock_step_cache.side_effect = [
        step21,  # 2.1
        nothing_step,  # 1.1
        jump_step(['sg4']),  # 1.2
        jump_step(['sg3']),  # 4.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
        nothing_step,  # 5.1
    ]

    context = Context({'a': 'b'})
    StepsRunner(get_while_pipeline(), context).run_step_groups(
        groups=['sg2'],
        success_group='sg5',
        failure_group=None)

    assert mock21.mock_calls == [call({'a': 'b', 'whileCounter': 1}),
                                 call({'a': 'b', 'whileCounter': 2})]
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg1.step1'),
                                          call('sg1.step2'),
                                          call('sg4.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2'),
                                          call('sg5.step1')]

    assert context == {'a': 'b', 'whileCounter': 2}


def get_for_pipeline():
    """Test pipeline for for."""
    return {
        'sg1': [
            'sg1.step1',
            'sg1.step2'
        ],
        'sg2': [
            {'name': 'sg2.step1',
             'foreach': ['one', 'two', 'three']
             },
            'sg2.step2'
        ],
        'sg3': [
            'sg3.step1',
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_jump_with_for(mock_step_cache):
    """Jump between different step groups in a foreach."""
    # Sequence: sg2 - sg2.1 x2 (JUMP)
    #           sg1 - sg1.1, sg 1.2 (JUMP)
    #           sg4 - sg4.1 (JUMP)
    #           sg3 - sg3.1, sg 3.2
    #           sg5 - sg5.1 (on_success)

    mock21 = DeepCopyMagicMock()

    def step21(context):
        mock21(context)
        if context['i'] == 'two':
            jump_step(['sg1'])(context)

    mock_step_cache.side_effect = [
        step21,  # 2.1
        nothing_step,  # 1.1
        jump_step(['sg4']),  # 1.2
        jump_step(['sg3']),  # 4.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
        nothing_step,  # 5.1
    ]

    context = Context({'a': 'b'})
    StepsRunner(get_for_pipeline(), context).run_step_groups(
        groups=['sg2'],
        success_group='sg5',
        failure_group=None)

    assert mock21.mock_calls == [call({'a': 'b', 'i': 'one'}),
                                 call({'a': 'b', 'i': 'two'})]
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg1.step1'),
                                          call('sg1.step2'),
                                          call('sg4.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2'),
                                          call('sg5.step1')]

    assert context == {'a': 'b', 'i': 'two'}
# endregion Jump


# region StopStepGroup
@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_step_group_with_success_handler(mock_step_cache):
    """Stop step group with success handler."""
    # Sequence: sg2 - sg2.1, sg2.2
    #           sg1 - sg1.1 STOP
    #           sg3 - sg3.1, sg3.2 - success handler
    def stop_step_group_step(context):
        raise StopStepGroup()

    mock_step_cache.side_effect = [
        nothing_step,  # 2.1
        nothing_step,  # 2.2
        stop_step_group_step,  # 1.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
    ]

    context = Context()
    StepsRunner(get_jump_pipeline(), context).run_step_groups(
        groups=['sg2', 'sg1'],
        success_group='sg3',
        failure_group=None)

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg1.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2')
                                          ]


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_step_group_with_jumps(mock_step_cache):
    """Stop Step Group with jumps."""
    # Sequence: sg2 - sg2.1 (JUMP)
    #           sg1 - sg1.1, sg 1.2 - STOP
    def stop_step_group_step(context):
        raise StopStepGroup()

    mock_step_cache.side_effect = [
        jump_step(['sg1']),  # 2.1
        nothing_step,  # 1.1
        stop_step_group_step  # 1.2
    ]

    context = Context()
    StepsRunner(get_jump_pipeline(), context).run_step_groups(
        groups=['sg2'],
        success_group=None,
        failure_group=None)

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg1.step1'),
                                          call('sg1.step2')]


def get_for_11_pipeline():
    """Test pipeline for for."""
    return {
        'sg1': [
            {'name': 'sg1.step1',
             'foreach': ['one', 'two', 'three']
             },
            'sg1.step2'
        ],
        'sg2': [
            'sg2.step1',
            'sg2.step2'
        ],
        'sg3': [
            'sg3.step1',
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_step_group_with_success_handler_for(mock_step_cache):
    """Stop step group with success handler in for loop."""
    # Sequence: sg2 - sg2.1, sg2.2
    #           sg1 - sg1.1 x2 STOP
    #           sg3 - sg3.1, sg3.2 - success handler
    mock11 = DeepCopyMagicMock()

    def step11(context):
        mock11(context)
        if context['i'] == 'two':
            raise StopStepGroup()

    mock_step_cache.side_effect = [
        nothing_step,  # 2.1
        nothing_step,  # 2.2
        step11,  # 1.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
    ]

    context = Context({'a': 'b'})
    StepsRunner(get_for_11_pipeline(), context).run_step_groups(
        groups=['sg2', 'sg1'],
        success_group='sg3',
        failure_group=None)

    assert mock11.mock_calls == [call({'a': 'b', 'i': 'one'}),
                                 call({'a': 'b', 'i': 'two'})]
    assert context == {'a': 'b', 'i': 'two'}

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg1.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2')
                                          ]


def get_while_11_pipeline():
    """Test pipeline for while."""
    return {
        'sg1': [
            {'name': 'sg1.step1',
             'while': {
                 'max': 3}
             },
            'sg1.step2'
        ],
        'sg2': [
            'sg2.step1',
            'sg2.step2'
        ],
        'sg3': [
            'sg3.step1',
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_stop_step_group_with_success_handler_while(mock_step_cache):
    """Stop step group with success handler in while loop."""
    # Sequence: sg2 - sg2.1, sg2.2
    #           sg1 - sg1.1 x2 STOP
    #           sg3 - sg3.1, sg3.2 - success handler
    mock11 = DeepCopyMagicMock()

    def step11(context):
        mock11(context)
        if context['whileCounter'] == 2:
            raise StopStepGroup()

    mock_step_cache.side_effect = [
        nothing_step,  # 2.1
        nothing_step,  # 2.2
        step11,  # 1.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
    ]

    context = Context({'a': 'b'})
    StepsRunner(get_while_11_pipeline(), context).run_step_groups(
        groups=['sg2', 'sg1'],
        success_group='sg3',
        failure_group=None)

    assert mock11.mock_calls == [call({'a': 'b', 'whileCounter': 1}),
                                 call({'a': 'b', 'whileCounter': 2})]
    assert context == {'a': 'b', 'whileCounter': 2}

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg2.step2'),
                                          call('sg1.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2')
                                          ]
# endregion StopStepGroup


# region Call
@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_call_with_success_handler(mock_step_cache):
    """Call between different step groups with success handler."""
    # Sequence: sg2 - sg2.1 (CALL)
    #           sg1 - sg1.1, sg 1.2 (CALL)
    #           sg4 - sg4.1 (CALL)
    #           sg3 - sg3.1, sg 3.2
    #           sg 4.2, sg2.2 (come back to call point)
    #           sg5 - sg5.1 (on_success)
    mock_step_cache.side_effect = [
        call_step(['sg1']),  # 2.1
        nothing_step,  # 1.1
        call_step(['sg4']),  # 1.2
        call_step(['sg3']),  # 4.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
        nothing_step,  # 4.2
        nothing_step,  # 2.2
        nothing_step,  # 5.1
    ]

    context = Context({'call': {'groups': 'sg1'}})

    pipeline = Pipeline('arb')
    steps_runner = StepsRunner(get_jump_pipeline(), context)
    pipeline.steps_runner = steps_runner

    with context.pipeline_scope(pipeline):
        steps_runner.run_step_groups(
            groups=['sg2'],
            success_group='sg5',
            failure_group=None)

    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg1.step1'),
                                          call('sg1.step2'),
                                          call('sg4.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2'),
                                          call('sg4.step2'),
                                          call('sg2.step2'),
                                          call('sg5.step1')]


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_call_with_failure_handler(mock_step_cache):
    """Call between different step groups with failure handler."""
    # Sequence: sg2 - sg2.1 (CALL)
    #           sg3 - sg3.1 (ERROR)
    #           sg4 - sg4.1, sg4.2 (failure handler)
    def err_step(context):
        raise ValueError('3.1')

    mock_step_cache.side_effect = [
        call_step(['sg3']),  # 2.1
        err_step,  # 3.1
        nothing_step,  # 4.1
        nothing_step,  # 4.2
    ]

    context = Context({'call': {'groups': 'sg3'}})

    pipeline = Pipeline('arb')
    steps_runner = StepsRunner(get_jump_pipeline(), context)
    pipeline.steps_runner = steps_runner

    with context.pipeline_scope(pipeline):
        with pytest.raises(ValueError) as err_info:
            steps_runner.run_step_groups(
                groups=['sg2', 'sg1'],
                success_group='sg5',
                failure_group='sg4')

    assert str(err_info.value) == '3.1'
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg3.step1'),
                                          call('sg4.step1'),
                                          call('sg4.step2')]


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_call_with_failure_handler_while(mock_step_cache):
    """Call between different step groups with failure handler."""
    # Sequence: sg2 - sg2.1 x2 (CALL)
    #           sg3 - sg3.1 (ERROR)
    #           sg4 - sg4.1, sg4.2 (failure handler)
    def err_step(context):
        raise ValueError('3.1')

    mock21 = DeepCopyMagicMock()

    def step21(context):
        mock21(context)
        if context['whileCounter'] == 2:
            context['call'] = "step21"
            call_step(['sg3'])(context)

    mock_step_cache.side_effect = [
        step21,  # 2.1
        err_step,  # 3.1
        nothing_step,  # 4.1
        nothing_step,  # 4.2
    ]

    context = Context({'a': 'b'})

    pipeline = Pipeline('arb')
    steps_runner = StepsRunner(get_while_pipeline(), context)
    pipeline.steps_runner = steps_runner

    with context.pipeline_scope(pipeline):
        with pytest.raises(ValueError) as err_info:
            steps_runner.run_step_groups(
                groups=['sg2', 'sg1'],
                success_group='sg5',
                failure_group='sg4')

    assert str(err_info.value) == '3.1'
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg3.step1'),
                                          call('sg4.step1'),
                                          call('sg4.step2')]

    assert mock21.mock_calls == [call({'a': 'b',
                                       'whileCounter': 1
                                       }),
                                 call({'a': 'b',
                                       'whileCounter': 2
                                       })]

    assert repr(context) == repr({'a': 'b',
                                  'whileCounter': 2,
                                  'call': 'arb',
                                  'runErrors': [
                                      {'name': 'ValueError',
                                       'description': '3.1',
                                       'customError': {},
                                       'line': None,
                                       'col': None,
                                       'step': 'sg3.step1',
                                       'exception': ValueError('3.1'),
                                       'swallowed': False}
                                  ]})


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_call_with_no_failure_handler(mock_step_cache):
    """Call between different step groups raising an error and no handler."""
    # Sequence: sg2 - sg2.1 (CALL)
    #           sg3 - sg3.1 (ERROR)
    #           sg4 - sg4.1, sg4.2 (failure handler)
    def err_step(context):
        raise ValueError('3.1')

    mock_step_cache.side_effect = [
        call_step(['sg3']),  # 2.1
        err_step,  # 3.1
        nothing_step,  # 4.1
        nothing_step,  # 4.2
    ]

    context = Context({'call': {'groups': 'sg3'}})

    pipeline = Pipeline('arb')
    steps_runner = StepsRunner(get_jump_pipeline(), context)
    pipeline.steps_runner = steps_runner

    with context.pipeline_scope(pipeline):
        with pytest.raises(ValueError) as err_info:
            steps_runner.run_step_groups(
                groups=['sg2', 'sg1'],
                success_group=None,
                failure_group=None)

    assert str(err_info.value) == '3.1'
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg3.step1')]


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_call_with_success_handler_for(mock_step_cache):
    """Call between different step groups with success handler in for."""
    # Sequence: sg2 - sg2.1 x2 (CALL)
    #           sg1 - sg1.1, sg 1.2 (CALL)
    #           sg4 - sg4.1 (CALL)
    #           sg3 - sg3.1, sg 3.2
    #           sg 4.2, (come back to call point)
    #           sg2.1 x1 (back in for iterator)
    #           sg2.2 (complete sg2)
    #           sg5 - sg5.1 (on_success)
    mock21 = DeepCopyMagicMock()

    def step21(context):
        mock21(context)
        if context['i'] == 'two':
            context['call'] = 'sg1'
            call_step(['sg1'])(context)

    def mutate_step(context):
        context['a'] = 'changed'

    def mutate_after_loop(context):
        context['a'] = 'after loop'

    mock_step_cache.side_effect = [
        step21,  # 2.1
        nothing_step,  # 1.1
        call_step(['sg4']),  # 1.2
        call_step(['sg3']),  # 4.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
        mutate_step,  # 4.2
        mutate_after_loop,  # 2.2
        nothing_step,  # 5.1
    ]

    context = Context({'a': 'b'})
    pipeline = Pipeline('arb')
    steps_runner = StepsRunner(get_for_pipeline(), context)
    pipeline.steps_runner = steps_runner

    with context.pipeline_scope(pipeline):
        steps_runner.run_step_groups(
            groups=['sg2'],
            success_group='sg5',
            failure_group=None)

    assert mock21.mock_calls == [call({'a': 'b',
                                       'i': 'one'}),
                                 call({'a': 'b',
                                       'i': 'two'}),
                                 call({'a': 'changed',
                                       'i': 'three',
                                       'call': 'arb'})]
    assert context == {'a': 'after loop', 'i': 'three', 'call': 'arb'}
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg1.step1'),
                                          call('sg1.step2'),
                                          call('sg4.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2'),
                                          call('sg4.step2'),
                                          call('sg2.step2'),
                                          call('sg5.step1')]


def get_retry_pipeline():
    """Test pipeline for retry."""
    return {
        'sg1': [
            'sg1.step1',
            'sg1.step2'
        ],
        'sg2': [
            {'name': 'sg2.step1',
             'retry': {
                 'max': 3}
             },
            'sg2.step2'
        ],
        'sg3': [
            'sg3.step1',
            'sg3.step2'
        ],
        'sg4': [
            'sg4.step1',
            'sg4.step2'
        ],
        'sg5': [
            'sg5.step1'
        ],
        'sg6': [
            'sg6.step1',
            'sg6.step2'
        ]
    }


@patch('pypyr.cache.stepcache.step_cache.get_step')
def test_call_with_success_handler_retry(mock_step_cache):
    """Call between different step groups with success handler in retry."""
    # Sequence: sg2 - sg2.1 x2 (CALL)
    #           sg1 - sg1.1, sg 1.2 (CALL)
    #           sg4 - sg4.1 (CALL)
    #           sg3 - sg3.1, sg 3.2
    #           sg 4.2, sg2.2 (come back to call point)
    #           sg5 - sg5.1 (on_success)
    mock21 = DeepCopyMagicMock()

    def step21(context):
        mock21(context)
        if context['retryCounter'] == 2:
            context['call'] = 'sg1'
            call_step(['sg1'])(context)
        else:
            raise ValueError(context['retryCounter'])

    mock_step_cache.side_effect = [
        step21,  # 2.1
        nothing_step,  # 1.1
        call_step(['sg4']),  # 1.2
        call_step(['sg3']),  # 4.1
        nothing_step,  # 3.1
        nothing_step,  # 3.2
        nothing_step,  # 4.2
        nothing_step,  # 2.2
        nothing_step,  # 5.1
    ]

    context = Context({'a': 'b'})
    pipeline = Pipeline('arb')
    steps_runner = StepsRunner(get_retry_pipeline(), context)
    pipeline.steps_runner = steps_runner

    with context.pipeline_scope(pipeline):
        steps_runner.run_step_groups(
            groups=['sg2'],
            success_group='sg5',
            failure_group=None)

    assert mock21.mock_calls == [call({'a': 'b',
                                       'retryCounter': 1}),
                                 call({'a': 'b',
                                       'retryCounter': 2})]
    assert context == {'a': 'b',
                       'retryCounter': 2,
                       'call': 'arb'}
    assert mock_step_cache.mock_calls == [call('sg2.step1'),
                                          call('sg1.step1'),
                                          call('sg1.step2'),
                                          call('sg4.step1'),
                                          call('sg3.step1'),
                                          call('sg3.step2'),
                                          call('sg4.step2'),
                                          call('sg2.step2'),
                                          call('sg5.step1')]

# endregion Call



"""safeshell.py unit tests."""
from pathlib import Path

import pytest

from pypyr.config import config
from pypyr.context import Context
from pypyr.errors import KeyNotInContextError
import pypyr.steps.safeshell

cmd_path = Path.cwd().joinpath('tests/testfiles/cmds')
is_windows = config.is_windows


def get_cmd(posix, win):
    """Return posix or win depending on current platform."""
    if is_windows:
        return str(cmd_path.joinpath(win))
    return posix


def test_safeshell_str_with_args():
    """Single command string works with string interpolation."""
    cmd = get_cmd('tests/testfiles/cmds/args.sh',
                  'args.bat')
    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd,
                       'cmd': '{d} {a} "{b}" {c}'})
    pypyr.steps.safeshell.run_step(context)

    assert 'cmdOut' not in context


def test_safeshell_dict_input_with_args():
    """Single command string works with multiple args."""
    cmd = get_cmd('tests/testfiles/cmds/args.sh',
                  'args.bat')
    context = Context({'a': 'one',
                       'b': 'two two',
                       'c': 'three',
                       'd': cmd,
                       'cmd': {
                           'run': '{d} {a} "{b}" {c}'}})
    pypyr.steps.safeshell.run_step(context)

    assert 'cmdOut' not in context


def test_empty_context_safeshell_throw():
    """Empty cmd in context should throw assert error."""
    with pytest.raises(KeyNotInContextError) as err_info:
        context = Context({'blah': 'blah blah'})
        pypyr.steps.safeshell.run_step(context)

    assert str(err_info.value) == ("context['cmd'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.cmd.")



"""fileformat.py unit tests."""
from unittest.mock import patch

import pytest

from pypyr.context import Context
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.fileformat as fileformat


def test_fileformat_no_inpath_raises():
    """None in path raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        fileformat.run_step(context)

    assert str(err_info.value) == (
        "context['fileFormat'] doesn't exist. "
        "It must exist for pypyr.steps.fileformat.")


def test_fileformat_empty_inpath_raises():
    """Empty in path raises."""
    context = Context({
        'fileFormat': {'in': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        fileformat.run_step(context)

    assert str(err_info.value) == ("context['fileFormat']['in'] must have a "
                                   "value for pypyr.steps.fileformat.")


def test_fileformat_pass_no_substitutions(fs):
    """Relative path to file should succeed."""
    payload = ('this is line 1\n'
               'this is line 2\n'
               'this is line 3\n'
               'this is line 4\n'
               'this !£$% * is line 5\n')

    in_path = './tests/testfiles/test.txt'
    fs.create_file(in_path, contents=payload)

    context = Context({
        'ok1': 'ov1',
        'fileFormat': {'in': in_path,
                       'out': './tests/testfiles/out/out.txt'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['ok1'] == 'ov1'
    assert context['fileFormat'] == {'in': in_path,
                                     'out': './tests/testfiles/out/out.txt'}

    with open('./tests/testfiles/out/out.txt') as outfile:
        outcontents = outfile.read()

    assert outcontents == payload


def test_fileformat_edit_no_substitutions(fs):
    """Relative path to file should succeed and edit in place when no out."""
    payload = ('this is line 1\n'
               'this is line 2\n'
               'this is line 3\n'
               'this is line 4\n'
               'this !£$% * is line 5\n')

    in_path = './tests/testfiles/out/edittest.txt'
    fs.create_file(in_path, contents=payload)
    context = Context({
        'ok1': 'ov1',
        'fileFormat': {'in': in_path}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['ok1'] == 'ov1'
    assert context['fileFormat'] == {
        'in': in_path}

    with open(in_path) as outfile:
        outcontents = outfile.read()

    assert outcontents == payload


def test_fileformat_pass_with_substitutions(fs):
    """Relative path to file should succeed."""
    payload = """this {k1} is line 1
this is line 2 {k2}
this is line 3
this {k3} is  {k4} line 4
this !£$% * is {k5} line 5
"""

    in_path = './tests/testfiles/testsubst.txt'
    fs.create_file(in_path, contents=payload)

    context = Context({
        'k1': 'v1',
        'k2': 'v2',
        'k3': 'v3',
        'k4': 'v4',
        'k5': 'v5',
        'fileFormat': {'in': './tests/testfiles/testsubst.txt',
                       'out': './tests/testfiles/out/outsubst.txt'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 6, "context should have 6 items"
    assert context['k1'] == 'v1'
    assert context['fileFormat'] == {
        'in': './tests/testfiles/testsubst.txt',
        'out': './tests/testfiles/out/outsubst.txt'}

    with open('./tests/testfiles/out/outsubst.txt') as outfile:
        outcontents = outfile.read()

    assert outcontents == """this v1 is line 1
this is line 2 v2
this is line 3
this v3 is  v4 line 4
this !£$% * is v5 line 5
"""


def test_fileformat_pass_with_path_substitutions(fs):
    """Relative path to file should succeed with path substitutions."""
    payload = """this {k1} is line 1
this is line 2 {k2}
this is line 3
this {k3} is  {k4} line 4
this !£$% * is {k5} line 5
"""

    in_path = './tests/testfiles/testsubst.txt'
    fs.create_file(in_path, contents=payload)

    context = Context({
        'k1': 'v1',
        'k2': 'v2',
        'k3': 'v3',
        'k4': 'v4',
        'k5': 'v5',
        'inFileName': 'testsubst',
        'outFileName': 'outsubst',
        'fileFormat': {'in': './tests/testfiles/{inFileName}.txt',
                       'out': './tests/testfiles/out/{outFileName}.txt'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 8, "context should have 8 items"
    assert context['k1'] == 'v1'
    assert context['fileFormat'] == {
        'in': './tests/testfiles/{inFileName}.txt',
        'out': './tests/testfiles/out/{outFileName}.txt'}

    with open('./tests/testfiles/out/outsubst.txt') as outfile:
        outcontents = outfile.read()

    assert outcontents == """this v1 is line 1
this is line 2 v2
this is line 3
this v3 is  v4 line 4
this !£$% * is v5 line 5
"""


@patch('pypyr.config.config.default_encoding', new='utf-16')
def test_fileformat_pass_with_encoding_from_config(fs):
    """Get encoding from config."""
    payload = """this {k1} is line 1
this is line 2 {k2}
this is line 3
this {k3} is  {k4} line 4
this !£$% * is {k5} line 5
"""

    in_path = './tests/testfiles/testsubst.txt'
    fs.create_file(in_path, contents=payload, encoding='utf-16')

    context = Context({
        'k1': 'v1',
        'k2': 'v2',
        'k3': 'v3',
        'k4': 'v4',
        'k5': 'v5',
        'enc': 'utf-16',
        'inFileName': 'testsubst',
        'outFileName': 'outsubst',
        'fileFormat': {'in': './tests/testfiles/{inFileName}.txt',
                       'out': './tests/testfiles/out/{outFileName}.txt'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 9, "context should have 9 items"
    assert context['k1'] == 'v1'
    assert context['fileFormat'] == {
        'in': './tests/testfiles/{inFileName}.txt',
        'out': './tests/testfiles/out/{outFileName}.txt'}

    with open('./tests/testfiles/out/outsubst.txt',
              encoding='utf-16') as outfile:
        outcontents = outfile.read()

    assert outcontents == """this v1 is line 1
this is line 2 v2
this is line 3
this v3 is  v4 line 4
this !£$% * is v5 line 5
"""


def test_fileformat_pass_with_encoding(fs):
    """Pass encoding with input."""
    payload = """this {k1} is line 1
this is line 2 {k2}
this is line 3
this {k3} is  {k4} line 4
this !£$% * is {k5} line 5
"""

    in_path = './tests/testfiles/testsubst.txt'
    fs.create_file(in_path, contents=payload, encoding='utf-16')

    context = Context({
        'k1': 'v1',
        'k2': 'v2',
        'k3': 'v3',
        'k4': 'v4',
        'k5': 'v5',
        'enc': 'utf-16',
        'inFileName': 'testsubst',
        'outFileName': 'outsubst',
        'fileFormat': {'in': './tests/testfiles/{inFileName}.txt',
                       'out': './tests/testfiles/out/{outFileName}.txt',
                       'encoding': '{enc}'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 9, "context should have 9 items"
    assert context['k1'] == 'v1'
    assert context['fileFormat'] == {
        'in': './tests/testfiles/{inFileName}.txt',
        'out': './tests/testfiles/out/{outFileName}.txt',
        'encoding': '{enc}'}

    with open('./tests/testfiles/out/outsubst.txt',
              encoding='utf-16') as outfile:
        outcontents = outfile.read()

    assert outcontents == """this v1 is line 1
this is line 2 v2
this is line 3
this v3 is  v4 line 4
this !£$% * is v5 line 5
"""


def test_fileformat_pass_with_encoding_in_to_out(fs):
    """Pass encoding with input and output."""
    payload = """this {k1} is line 1
this is line 2 {k2}
this is line 3
this {k3} is  {k4} line 4
this !£$% * is {k5} line 5
"""

    in_path = './tests/testfiles/testsubst.txt'
    fs.create_file(in_path, contents=payload, encoding='utf-16')

    context = Context({
        'k1': 'v1',
        'k2': 'v2',
        'k3': 'v3',
        'k4': 'v4',
        'k5': 'v5',
        'encIn': 'utf-16',
        'encOut': 'utf-32',
        'inFileName': 'testsubst',
        'outFileName': 'outsubst',
        'fileFormat': {'in': './tests/testfiles/{inFileName}.txt',
                       'out': './tests/testfiles/out/{outFileName}.txt',
                       'encodingIn': '{encIn}',
                       'encodingOut': '{encOut}'}})

    fileformat.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 10, "context should have 10 items"
    assert context['k1'] == 'v1'
    assert context['fileFormat'] == {
        'in': './tests/testfiles/{inFileName}.txt',
        'out': './tests/testfiles/out/{outFileName}.txt',
        'encodingIn': '{encIn}',
        'encodingOut': '{encOut}'}

    with open('./tests/testfiles/out/outsubst.txt',
              encoding='utf-32') as outfile:
        outcontents = outfile.read()

    assert outcontents == """this v1 is line 1
this is line 2 v2
this is line 3
this v3 is  v4 line 4
this !£$% * is v5 line 5
"""



"""asserts.py unit tests."""
import pytest
from pypyr.errors import (ContextError,
                          KeyNotInContextError,
                          KeyInContextHasNoValueError)
from pypyr.utils import asserts

# region assert_key_exists


def test_assert_key_exists_none():
    """Key value is None."""
    asserts.assert_key_exists({None: 'b'}, None, 'caller')

    with pytest.raises(KeyNotInContextError) as err:
        asserts.assert_key_exists({'a': 'b'}, None, 'caller')

    assert str(err.value) == (
        "context[None] doesn't exist. It must exist for caller.")


def test_assert_key_exists_int():
    """Key value is an int."""
    asserts.assert_key_exists({1: 'b'}, 1, 'caller')

    with pytest.raises(KeyNotInContextError) as err:
        asserts.assert_key_exists({'a': 'b'}, 1, 'caller')

    assert str(err.value) == (
        "context[1] doesn't exist. It must exist for caller.")


def test_assert_key_exists_with_parent():
    """Parent gives correct error message."""
    asserts.assert_key_exists({'a': 'b'}, 'a', 'caller', 'parent')

    with pytest.raises(KeyNotInContextError) as err:
        asserts.assert_key_exists({'a': 'b'}, 'key', 'caller', 'parent')

    assert str(err.value) == (
        "context['parent']['key'] doesn't exist. It must exist for caller.")


def test_assert_key_has_value_object_not_iterable():
    """Object is None should raise."""
    with pytest.raises(ContextError) as err:
        asserts.assert_key_has_value(obj=1,
                                     key=2,
                                     caller='arb caller',
                                     parent=3)

    assert str(err.value) == (
        "context[3] must exist, be iterable and contain 2 for "
        "arb caller. argument of type 'int' is not iterable")


def test_assert_key_has_value_object_not_iterable_no_parent():
    """Object is None should raise with no parent."""
    with pytest.raises(ContextError) as err:
        asserts.assert_key_has_value(obj=1,
                                     key=2,
                                     caller='arb caller')

    assert str(err.value) == (
        "context[2] must exist and be iterable for arb caller. argument of "
        "type 'int' is not iterable")

# endregion assert_key_exists

# region assert_key_has_value


def test_assert_key_has_value_passes():
    """Success case."""
    asserts.assert_key_has_value(obj={'k1': 'v1'},
                                 key='k1',
                                 caller='arb caller',
                                 parent='parent name')


def test_assert_key_value_empty():
    """Not a truthy."""
    asserts.assert_key_has_value(obj={'k1': ''},
                                 key='k1',
                                 caller='arb caller',
                                 parent='parent name')


def test_assert_key_has_value_none():
    """Key value is none."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        asserts.assert_key_has_value(obj={'k1': None},
                                     key='k1',
                                     caller='arb caller',
                                     parent='parent name')

    assert str(err.value) == ("context['parent name']['k1'] must have a value "
                              "for arb caller.")


def test_assert_key_has_value_none_no_parent():
    """Key value is none and no parent."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        asserts.assert_key_has_value(obj={'k1': None},
                                     key='k1',
                                     caller='arb caller')

    assert str(err.value) == "context['k1'] must have a value for arb caller."


def test_assert_key_has_value_key_not_there():
    """Key not in object."""
    with pytest.raises(KeyNotInContextError) as err:
        asserts.assert_key_has_value(obj={'k1': None},
                                     key='k2',
                                     caller='arb caller',
                                     parent='parent name')

    assert str(err.value) == ("context['parent name']['k2'] doesn't exist. "
                              "It must exist for arb caller.")


def test_assert_key_has_value_key_not_there_no_parent():
    """Key not in object and no parent."""
    with pytest.raises(KeyNotInContextError) as err:
        asserts.assert_key_has_value(obj={'k1': None},
                                     key='k2',
                                     caller='arb caller')

    assert str(err.value) == ("context['k2'] doesn't exist. "
                              "It must exist for arb caller.")


def test_assert_key_has_value_object_none():
    """Object is None should raise."""
    with pytest.raises(ContextError) as err:
        asserts.assert_key_has_value(obj=None,
                                     key='k1',
                                     caller='arb caller',
                                     parent='parent name')

    assert str(err.value) == (
        "context['parent name'] must exist, be iterable and contain 'k1' for "
        "arb caller. argument of type 'NoneType' is not iterable")


def test_assert_key_has_value_object_none_no_parent():
    """Object is None should raise with no parent."""
    with pytest.raises(ContextError) as err:
        asserts.assert_key_has_value(obj=None,
                                     key='k1',
                                     caller='arb caller')

    assert str(err.value) == (
        "context['k1'] must exist and be iterable for arb caller. argument of "
        "type 'NoneType' is not iterable")
# endregion assert_key_has_value

# region assert_key_is_truthy


def test_assert_key_is_truthy_passes():
    """Success case."""
    asserts.assert_key_is_truthy(obj={'k1': 'v1'},
                                 key='k1',
                                 caller='arb caller',
                                 parent='parent name')


def test_assert_key_is_truthy_value_empty():
    """Is a truthy."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        asserts.assert_key_is_truthy(obj={'k1': ''},
                                     key='k1',
                                     caller='arb caller',
                                     parent='parent name')

    assert str(err.value) == ("context['parent name']['k1'] must have a value "
                              "for arb caller.")


def test_assert_key_is_truthy_none():
    """Key value is none."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        asserts.assert_key_is_truthy(obj={'k1': None},
                                     key='k1',
                                     caller='arb caller',
                                     parent='parent name')

    assert str(err.value) == ("context['parent name']['k1'] must have a value "
                              "for arb caller.")


def test_assert_key_is_truthy_none_no_parent():
    """Key value is none and no parent."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        asserts.assert_key_is_truthy(obj={'k1': None},
                                     key='k1',
                                     caller='arb caller')

    assert str(err.value) == "context['k1'] must have a value for arb caller."


def test_assert_key_is_truthy_key_not_there():
    """Key not in object."""
    with pytest.raises(KeyNotInContextError) as err:
        asserts.assert_key_is_truthy(obj={'k1': None},
                                     key='k2',
                                     caller='arb caller',
                                     parent='parent name')

    assert str(err.value) == ("context['parent name']['k2'] doesn't exist. "
                              "It must exist for arb caller.")


def test_assert_key_is_truthy_key_not_there_no_parent():
    """Key not in object and no parent."""
    with pytest.raises(KeyNotInContextError) as err:
        asserts.assert_key_is_truthy(obj={'k1': None},
                                     key='k2',
                                     caller='arb caller')

    assert str(err.value) == ("context['k2'] doesn't exist. "
                              "It must exist for arb caller.")


def test_assert_key_is_truthy_object_none():
    """Object is None should raise."""
    with pytest.raises(ContextError) as err:
        asserts.assert_key_is_truthy(obj=None,
                                     key='k1',
                                     caller='arb caller',
                                     parent='parent name')

    assert str(err.value) == (
        "context['parent name'] must exist, be iterable and contain 'k1' for "
        "arb caller. argument of type 'NoneType' is not iterable")


def test_assert_key_is_truthy_object_none_no_parent():
    """Object is None should raise with no parent."""
    with pytest.raises(ContextError) as err:
        asserts.assert_key_is_truthy(obj=None,
                                     key='k1',
                                     caller='arb caller')

    assert str(err.value) == (
        "context['k1'] must exist and be iterable for arb caller. argument of "
        "type 'NoneType' is not iterable")
# endregion assert_key_is_truthy

# region assert_keys_are_truthy


def test_assert_keys_are_truthy():
    """Multiple keys on truthy check."""
    asserts.assert_keys_are_truthy(obj={'k1': 'v1', 'k2': 'v2', 'k3': 'v3'},
                                   keys=('k1', 'k3'),
                                   caller='arb caller',
                                   parent='parent name')


def test_assert_keys_are_truthy_none():
    """Key value is none for one of the keys raise error."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        asserts.assert_keys_are_truthy(obj={'k1': 'v1',
                                            'k2': None,
                                            'k3': 'v3'},
                                       keys=['k1', 'k3', 'k2'],
                                       caller='arb caller',
                                       parent='parent name')

    assert str(err.value) == ("context['parent name']['k2'] must have a value "
                              "for arb caller.")


def test_assert_keys_are_truthy_empty():
    """Key value is empty for one of the keys raise error."""
    with pytest.raises(KeyInContextHasNoValueError) as err:
        asserts.assert_keys_are_truthy(obj={'k1': 'v1',
                                            'k2': '',
                                            'k3': 'v3'},
                                       keys=['k1', 'k3', 'k2'],
                                       caller='arb caller',
                                       parent='parent name')

    assert str(err.value) == ("context['parent name']['k2'] must have a value "
                              "for arb caller.")
# endregion assert_keys_are_truthy



"""pypyr step that writes payload out to a toml file."""
import logging
from pathlib import Path

from pypyr.errors import KeyInContextHasNoValueError
import pypyr.toml as toml
from pypyr.utils.asserts import assert_key_has_value

logger = logging.getLogger(__name__)

sentinel = object()


def run_step(context):
    """Write payload out to toml file.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context keys expected:
                - fileWriteToml
                    - path. mandatory. path-like. Write output file to
                      here. Will create directories in path for you.
                    - payload. optional. Write this key to output file. If not
                      specified, output entire context.

    If you do specify a payload, it cannot be None or empty.

    Note that TOML is always in utf-8, per TOML spec.

    Returns:
        None.

    Raises:
        pypyr.errors.KeyNotInContextError: fileWriteToml or
            fileWriteToml.path missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fileWriteToml or
            fileWriteToml.path exists but is None.

    """
    logger.debug("started")

    input_key_name = 'fileWriteToml'
    context.assert_key_has_value(input_key_name, __name__)

    input_context = context.get_formatted(input_key_name)
    assert_key_has_value(obj=input_context,
                         key='path',
                         caller=__name__,
                         parent=input_key_name)

    out_path = Path(input_context['path'])
    # doing it like this to safeguard against accidentally dumping all context
    # with potentially sensitive values in it to disk if payload exists but is
    # None.
    payload = input_context.get('payload', sentinel)

    if payload is sentinel:
        payload = context.get_formatted_value(context)
    else:
        if not payload:
            raise KeyInContextHasNoValueError(
                "payload must have a value to write to output TOML document.")

        payload = input_context['payload']

    logger.debug("opening destination file for writing: %s", out_path)

    out_path.parent.mkdir(parents=True, exist_ok=True)

    toml.write_file(out_path, payload)

    logger.info("formatted context content and wrote to %s", out_path)
    logger.debug("done")



"""pathcheck.py unit tests."""
import logging
import pytest
from unittest.mock import call, patch
from pypyr.context import Context
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.pathcheck as pathchecker
from tests.common.utils import patch_logger


def test_pathcheck_no_input_raises():
    """No input context raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        pathchecker.run_step(context)

    assert str(err_info.value) == ("context['pathCheck'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.pathcheck.")


def test_pathcheck_empty_path_raises():
    """Empty path check raises."""
    context = Context({
        'pathCheck': None})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        pathchecker.run_step(context)

    assert str(err_info.value) == ("context['pathCheck'] must have a "
                                   "value for pypyr.steps.pathcheck.")


def test_pathcheck_empty_string_path_raises():
    """Empty string path check raises."""
    context = Context({
        'pathCheck': ''})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        pathchecker.run_step(context)

    assert str(err_info.value) == ("context['pathCheck'] must have a "
                                   "value for pypyr.steps.pathcheck.")


def test_pathcheck_empty_path_list_raises():
    """Empty path check list raises."""
    context = Context({
        'pathCheck': []})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        pathchecker.run_step(context)

    assert str(err_info.value) == ("context['pathCheck'] must have a "
                                   "value for pypyr.steps.pathcheck.")


@patch('pypyr.utils.filesystem.get_glob')
def test_pathcheck_single(mock_glob):
    """Single path ok."""
    context = Context({
        'ok1': 'ov1',
        'pathCheck': './arb/x'})

    mock_glob.return_value = ['./foundfile']

    with patch_logger(
            'pypyr.steps.pathcheck', logging.INFO) as mock_logger_info:
        pathchecker.run_step(context)

    mock_logger_info.assert_called_once_with('checked 1 path(s) and found 1')
    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['pathCheck'] == './arb/x'
    assert context["pathCheckOut"] == {'./arb/x': {
        'exists': True,
        'count': 1,
        'found': ['./foundfile']
    }}

    mock_glob.assert_called_once_with('./arb/x')


@patch('pypyr.utils.filesystem.get_glob')
def test_pathcheck_single_not_found(mock_glob):
    """Single path ok on not found."""
    context = Context({
        'ok1': 'ov1',
        'pathCheck': './arb/x'})

    mock_glob.return_value = []

    with patch_logger(
            'pypyr.steps.pathcheck', logging.INFO) as mock_logger_info:
        pathchecker.run_step(context)

    mock_logger_info.assert_called_once_with('checked 1 path(s) and found 0')

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['pathCheck'] == './arb/x'
    assert context["pathCheckOut"] == {'./arb/x': {
        'exists': False,
        'count': 0,
        'found': []
    }}

    mock_glob.assert_called_once_with('./arb/x')


@patch('pypyr.utils.filesystem.get_glob')
def test_pathcheck_single_with_formatting(mock_glob):
    """Single path ok with string formatting."""
    context = Context({
        'ok1': 'ov1',
        'pathCheck': './{ok1}/x'})

    mock_glob.return_value = ['./foundfile']

    pathchecker.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['pathCheck'] == './{ok1}/x'
    assert context["pathCheckOut"] == {'./{ok1}/x': {
        'exists': True,
        'count': 1,
        'found': ['./foundfile']
    }}

    mock_glob.assert_called_once_with('./ov1/x')


@patch('pypyr.utils.filesystem.get_glob')
def test_pathcheck_list(mock_glob):
    """Multiple paths ok with some returning no match."""
    context = Context({
        'ok1': 'ov1',
        'pathCheck': ['./arb/x', './arb/y', './arb/z']})

    mock_glob.side_effect = [
        ['./f1.1'],
        ['./f2.1', './f2.2', './f2.3'],
        []
    ]

    with patch_logger(
            'pypyr.steps.pathcheck', logging.INFO) as mock_logger_info:
        pathchecker.run_step(context)

    mock_logger_info.assert_called_once_with('checked 3 path(s) and found 4')

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['pathCheck'] == ['./arb/x', './arb/y', './arb/z']
    assert context["pathCheckOut"] == {
        './arb/x': {
            'exists': True,
            'count': 1,
            'found': ['./f1.1']
        },
        './arb/y': {
            'exists': True,
            'count': 3,
            'found': ['./f2.1', './f2.2', './f2.3']
        },
        './arb/z': {
            'exists': False,
            'count': 0,
            'found': []
        },
    }

    assert mock_glob.mock_calls == [call('./arb/x'),
                                    call('./arb/y'),
                                    call('./arb/z')]


@patch('pypyr.utils.filesystem.get_glob')
def test_pathcheck_list_none(mock_glob):
    """Multiple paths ok with none returning match."""
    context = Context({
        'ok1': 'ov1',
        'pathCheck': ['./{ok1}/x', './arb/{ok1}', '{ok1}/arb/z']})

    mock_glob.side_effect = [
        [],
        [],
        []
    ]

    with patch_logger(
            'pypyr.steps.pathcheck', logging.INFO) as mock_logger_info:
        pathchecker.run_step(context)

    mock_logger_info.assert_called_once_with('checked 3 path(s) and found 0')

    assert context, "context shouldn't be None"
    assert len(context) == 3, "context should have 3 items"
    assert context['ok1'] == 'ov1'
    assert context['pathCheck'] == ['./{ok1}/x', './arb/{ok1}', '{ok1}/arb/z']
    assert context["pathCheckOut"] == {
        './{ok1}/x': {
            'exists': False,
            'count': 0,
            'found': []
        },
        './arb/{ok1}': {
            'exists': False,
            'count': 0,
            'found': []
        },
        '{ok1}/arb/z': {
            'exists': False,
            'count': 0,
            'found': []
        },
    }

    assert mock_glob.mock_calls == [call('./ov1/x'),
                                    call('./arb/ov1'),
                                    call('ov1/arb/z')]



"""jsonparse.py unit tests."""
from pypyr.context import Context
from pypyr.dsl import SicString
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.jsonparse as jsonparse
import pytest


def test_jsonparse_no_jsonparse_raises():
    """No jsonparse input raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        jsonparse.run_step(context)

    assert str(err_info.value) == ("context['jsonParse'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.jsonparse.")


def test_jsonparse_no_json_raises():
    """No json raises."""
    context = Context({
        'jsonParse': {
            'a': 'b'}})

    with pytest.raises(KeyNotInContextError) as err_info:
        jsonparse.run_step(context)

    assert str(err_info.value) == (
        "context['jsonParse']['json'] doesn't exist. It must exist for "
        "pypyr.steps.jsonparse.")


def test_jsonparse_with_formatting():
    """Json with a root scalar works with out key."""
    context = Context({
        'k1': 'out',
        'k2': '{"a": "b", "c": "d"}',
        'jsonParse': {
            'json': '{k2:ff}',
            'key': '{k1}'}})

    jsonparse.run_step(context)

    assert context['out'] == {'a': 'b', 'c': 'd'}


def test_jsonparse_scalar_with_key():
    """Json with a root scalar works with out key."""
    context = Context({
        'ok1': 'ov1',
        'jsonParse': {
            'json': '1',
            'key': 'out'}})

    jsonparse.run_step(context)

    assert context['out'] == 1


def test_jsonparse_scalar_with_key_null():
    """Json with a root scalar works with out key and null value."""
    context = Context({
        'ok1': 'ov1',
        'jsonParse': {
            'json': 'null',
            'key': 'out'}})

    jsonparse.run_step(context)

    assert context['out'] is None


def test_jsonparse_scalar_with_key_empty():
    """Json with a root scalar works with out key and empty value."""
    context = Context({
        'ok1': 'ov1',
        'jsonParse': {
            'json': '',
            'key': 'out'}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        jsonparse.run_step(context)

    assert str(err_info.value) == (
        'jsonParse.json exists but is empty. It should be a valid json '
        'string for pypyr.steps.jsonparse. For example: '
        '\'{"key1": "value1", "key2": "value2"}\'')


def test_jsonparse_scalar_no_key_fails():
    """Json describing a scalar rather than a dict should fail if no outkey."""
    context = Context({
        'ok1': 'ov1',
        'jsonParse': {
            'json': '1'}})

    with pytest.raises(TypeError) as err_info:
        jsonparse.run_step(context)

    assert str(err_info.value) == (
        'json input should describe an object at the top level when '
        'jsonParse.key isn\'t specified. You should have something like '
        '\'{"key1": "value1", "key2": "value2"}\' in the json top-level, not '
        '["value1", "value2"]')


def test_jsonparse_map_no_key():
    """Json with a root mapping works with no key."""
    context = Context({
        'ok1': 'ov1',
        'jsonParse': {
            'json': SicString('{"a": "b", "c": "d"}')}})

    jsonparse.run_step(context)

    assert context == {
        'ok1': 'ov1',
        'jsonParse': {
            'json': SicString('{"a": "b", "c": "d"}')},
        'a': 'b',
        'c': 'd'}


def test_jsonparse_map_with_key():
    """Json with a root mapping works with out key."""
    context = Context({
        'ok1': 'ov1',
        'jsonParse': {
            'json': SicString('{"a": "b", "c": "d"}'),
            'key': 'out'}})

    jsonparse.run_step(context)

    assert context['out'] == {'a': 'b', 'c': 'd'}


def test_jsonparse_list_with_key():
    """Json with a root list works with out key."""
    context = Context({
        'ok1': 'ov1',
        'jsonParse': {
            'json': '[1, 2, 3]',
            'key': 'out'}})

    jsonparse.run_step(context)

    assert context['out'] == [1, 2, 3]


def test_jsonparse_list_fails():
    """Json describing a list rather than a dict should fail if no outkey."""
    context = Context({
        'ok1': 'ov1',
        'jsonParse': {
            'json': '[1, 2, 3]'}})

    with pytest.raises(TypeError) as err_info:
        jsonparse.run_step(context)

    assert str(err_info.value) == (
        'json input should describe an object at the top level when '
        'jsonParse.key isn\'t specified. You should have something like '
        '\'{"key1": "value1", "key2": "value2"}\' in the json top-level, not '
        '["value1", "value2"]')



"""fileformat.py unit tests."""
import pytest

from pypyr.context import Context
from pypyr.errors import KeyInContextHasNoValueError, KeyNotInContextError
import pypyr.steps.filereplace as filereplace


# region arg validation
def test_filereplace_no_inpath_raises():
    """None in path raises."""
    context = Context({
        'k1': 'v1'})

    with pytest.raises(KeyNotInContextError) as err_info:
        filereplace.run_step(context)

    assert str(err_info.value) == (
        "context['fileReplace'] doesn't exist. "
        "It must exist for pypyr.steps.filereplace.")


def test_filereplace_empty_inpath_raises():
    """Empty in path raises."""
    context = Context({
        'fileReplace': {'in': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filereplace.run_step(context)

    assert str(err_info.value) == ("context['fileReplace']['in'] must have a "
                                   "value for pypyr.steps.filereplace.")


def test_filereplace_no_replacepairs_raises():
    """None replacepairs raises."""
    context = Context({
        'fileReplace': {'in': 'blah',
                        'out': 'blah',
                        'k1': 'v1'}})

    with pytest.raises(KeyNotInContextError) as err_info:
        filereplace.run_step(context)

    assert str(err_info.value) == ("context['fileReplace']['replacePairs'] "
                                   "doesn't exist. It must exist for "
                                   "pypyr.steps.filereplace.")


def test_filereplace_empty_replacepairs_raises():
    """Empty in path raises."""
    context = Context({
        'fileReplace': {'in': 'blah',
                        'out': 'blah',
                        'replacePairs': None}})

    with pytest.raises(KeyInContextHasNoValueError) as err_info:
        filereplace.run_step(context)

    assert str(err_info.value) == (
        "context['fileReplace']['replacePairs'] must have "
        "a value for pypyr.steps.filereplace.")

# endregion arg validation

# region run_step


def test_filereplace_pass_no_matches(fs):
    """Relative path to file should succeed."""
    payload = ('this is line 1\n'
               'this is line 2\n'
               'this is line 3\n'
               'this is line 4\n'
               'this !£$% * is line 5\n')

    in_path = './tests/testfiles/test.txt'
    fs.create_file(in_path, contents=payload)

    context = Context({
        'ok1': 'ov1',
        'fileReplace': {'in': in_path,
                        'out': './tests/testfiles/out/outreplace.txt',
                        'replacePairs': {
                            'XXXXX': 'doesnt exist',
                            'YYYYY': 'doesnt exist either'
                        }}})

    filereplace.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['ok1'] == 'ov1'
    assert context['fileReplace'] == {
        'in': './tests/testfiles/test.txt',
        'out': './tests/testfiles/out/outreplace.txt',
        'replacePairs': {
            'XXXXX': 'doesnt exist',
            'YYYYY': 'doesnt exist either'
        }}

    with open('./tests/testfiles/out/outreplace.txt') as outfile:
        outcontents = outfile.read()

    assert outcontents == payload


def test_filereplace_pass_with_replacements(fs):
    """Relative path to file should succeed.

    Strictly speaking not a unit test.
    """
    payload = ('this {k1} X1 is line 1\n'
               'this is line 2 REPLACEME2\n'
               'this is line 3\n'
               'this rm3 RM3 is  RM4 line 4\n'
               'this !£$% * is rm5 line 5\n')

    in_path = './tests/testfiles/testreplace.txt'
    fs.create_file(in_path, contents=payload)

    context = Context({
        'k1': 'X1',
        'fileReplace': {'in': in_path,
                        'out': './tests/testfiles/out/outreplace.txt',
                        'replacePairs': {
                            '{k1}': 'v1',
                            'REPLACEME2': 'v2',
                            'RM3': 'v3',
                            'RM4': 'v4',
                            'rm5': 'v5',
                        }}})

    filereplace.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 2, "context should have 2 items"
    assert context['k1'] == 'X1'
    assert context['fileReplace'] == {
        'in': in_path,
        'out': './tests/testfiles/out/outreplace.txt',
        'replacePairs': {
            '{k1}': 'v1',
            'REPLACEME2': 'v2',
            'RM3': 'v3',
            'RM4': 'v4',
            'rm5': 'v5',
        }}
    with open('./tests/testfiles/out/outreplace.txt') as outfile:
        outcontents = list(outfile)

    assert outcontents[0] == "this {k1} v1 is line 1\n"
    assert outcontents[1] == "this is line 2 v2\n"
    assert outcontents[2] == "this is line 3\n"
    assert outcontents[3] == "this rm3 v3 is  v4 line 4\n"
    assert outcontents[4] == "this !£$% * is v5 line 5\n"


def test_filereplace_pass_with_substitutions(fs):
    """Relative path to file should succeed with input substitutions."""
    payload = ('this {k1} X1 is line 1\n'
               'this is line 2 REPLACEME2\n'
               'this is line 3\n'
               'this rm3 RM3 is  RM4 line 4\n'
               'this !£$% * is rm5 line 5\n')

    in_path = './tests/testfiles/testreplace.txt'
    fs.create_file(in_path, contents=payload, encoding='utf-16')

    context = Context({
        'k1': 'X1',
        'inFile': 'testreplace',
        'outFile': 'outreplace',
        'enc': 'utf-16',
        'fileReplace': {'in': './tests/testfiles/{inFile}.txt',
                        'out': './tests/testfiles/out/{outFile}.txt',
                        'encoding': '{enc}',
                        'replacePairs': {
                            '{k1}': 'v1',
                            'REPLACEME2': 'v2',
                            'RM3': 'v3',
                            'RM4': 'v4',
                            'rm5': 'v5'}}})

    filereplace.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['k1'] == 'X1'
    assert context['fileReplace'] == {
        'in': './tests/testfiles/{inFile}.txt',
        'out': './tests/testfiles/out/{outFile}.txt',
        'encoding': '{enc}',
        'replacePairs': {
            '{k1}': 'v1',
            'REPLACEME2': 'v2',
            'RM3': 'v3',
            'RM4': 'v4',
            'rm5': 'v5'}}

    with open('./tests/testfiles/out/outreplace.txt',
              encoding='utf-16') as file:
        outcontent = file.read()

    expected = ("this {k1} v1 is line 1\n"
                "this is line 2 v2\n"
                "this is line 3\n"
                "this rm3 v3 is  v4 line 4\n"
                "this !£$% * is v5 line 5\n")

    assert outcontent == expected


def test_filereplace_pass_out_is_dir(fs):
    """Relative path to file should write same filename if out is dir."""
    payload = ('this {k1} X1 is line 1\n'
               'this is line 2 REPLACEME2\n'
               'this is line 3\n'
               'this rm3 RM3 is  RM4 line 4\n'
               'this !£$% * is rm5 line 5\n')

    in_path = '/testreplace.txt'
    fs.create_file(in_path, contents=payload)

    context = Context({
        'k1': 'X1',
        'inFile': 'testreplace',
        'outDir': '/out/',
        'enc': 'utf-16',
        'fileReplace': {'in': '/{inFile}.txt',
                        'out': '{outDir}',
                        'encodingOut': '{enc}',
                        'replacePairs': {
                            '{k1}': 'v1',
                            'REPLACEME2': 'v2',
                            'RM3': 'v3',
                            'RM4': 'v4',
                            'rm5': 'v5'}}})

    filereplace.run_step(context)

    assert context, "context shouldn't be None"
    assert len(context) == 5, "context should have 5 items"
    assert context['k1'] == 'X1'
    assert context['fileReplace'] == {
        'in': '/{inFile}.txt',
        'out': '{outDir}',
        'encodingOut': '{enc}',
        'replacePairs': {
            '{k1}': 'v1',
            'REPLACEME2': 'v2',
            'RM3': 'v3',
            'RM4': 'v4',
            'rm5': 'v5'}}

    with open('/out/testreplace.txt', encoding='utf-16') as file:
        outcontent = file.read()

    expected = ("this {k1} v1 is line 1\n"
                "this is line 2 v2\n"
                "this is line 3\n"
                "this rm3 v3 is  v4 line 4\n"
                "this !£$% * is v5 line 5\n")

    assert outcontent == expected


def test_filereplace_pass_out_encoding_in_to_out(fs):
    """Change encoding from in to out."""
    payload = ('this {k1} X1 is line 1\n'
               'this is line 2 REPLACEME2\n'
               'this is line 3\n'
               'this rm3 RM3 is  RM4 line 4\n'
               'this !£$% * is rm5 line 5\n')

    in_path = '/testreplace.txt'
    fs.create_file(in_path, contents=payload, encoding='utf-32')

    context = Context({
        'k1': 'X1',
        'inFile': 'testreplace',
        'outDir': '/out/',
        'encIn': 'utf-32',
        'encOut': 'utf-16',
        'fileReplace': {'in': '/{inFile}.txt',
                        'out': '{outDir}',
                        'encodingIn': '{encIn}',
                        'encodingOut': '{encOut}',
                        'replacePairs': {
                            '{k1}': 'v1',
                            'REPLACEME2': 'v2',
                            'RM3': 'v3',
                            'RM4': 'v4',
                            'rm5': 'v5'}}})

    filereplace.run_step(context)

    with open('/out/testreplace.txt', encoding='utf-16') as file:
        outcontent = file.read()

    expected = ("this {k1} v1 is line 1\n"
                "this is line 2 v2\n"
                "this is line 3\n"
                "this rm3 v3 is  v4 line 4\n"
                "this !£$% * is v5 line 5\n")

    assert outcontent == expected
# endregion run_step



"""configvars.py unit tests."""
from pypyr.context import Context
from pypyr.steps import configvars


def test_configvars_empty_vars():
    """Do nothing when config.vars is empty."""
    context = Context()
    configvars.run_step(context)
    assert context == {}


def test_configvars_with_vars_to_empty_context(monkeypatch):
    """Copy to empty context when config.vars has values."""
    monkeypatch.setattr('pypyr.config.config.vars', {'a': 'b', 'c': 'd'})

    context = Context()
    configvars.run_step(context)
    assert context == {'a': 'b', 'c': 'd'}


def test_configvars_with_vars_to_context(monkeypatch):
    """Copy to existing context when config.vars has values."""
    monkeypatch.setattr('pypyr.config.config.vars', {'a': 'b', 'c': 'd'})

    context = Context({'og': 1, 'og2': 'v2'})
    configvars.run_step(context)
    assert context == {'og': 1, 'og2': 'v2', 'a': 'b', 'c': 'd'}



"""stoppipeline.py unit tests."""
import pytest
from pypyr.context import Context
from pypyr.errors import StopPipeline
import pypyr.steps.stoppipeline


def test_step_stoppipeline():
    """Stop raises stop."""
    with pytest.raises(StopPipeline):
        pypyr.steps.stoppipeline.run_step({})


def test_step_stoppipeline_context_same():
    """Stop doesn't nuke context."""
    context = Context({'test': 'value1'})
    with pytest.raises(StopPipeline):
        pypyr.steps.stoppipeline.run_step(context)
    assert context['test'] == 'value1', "context not returned from step."



"""Tests for pypyr.steps.cmds that touch the filesystem.

Can't use pyfakefs for these because the subprocess lib uses C libs to hit the
fs rather than the patch-able Python files access.
"""
from pathlib import Path
import tempfile
from unittest import TestCase

import pytest

from pypyr.config import config
from pypyr.context import Context
from pypyr.errors import MultiError, SubprocessError
import pypyr.steps.cmds

is_windows = config.is_windows


@pytest.fixture
def temp_dir():
    """Make tmp dir in testfiles/out. Yields pathlib.Path."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        yield Path(tmp_dir)


def get_cmd(posix, win):
    """Return posix or win depending on current platform."""
    return win if is_windows else posix


def test_async_cmds_stderr_to_stdout(temp_dir):
    """Redirect stderr to stdout."""
    cmd1 = get_cmd('echo one',
                   r'tests\testfiles\cmds\echo.bat one')

    cmd2 = get_cmd('tests/testfiles/cmds/exitwitherr.sh',
                   r'tests\testfiles\cmds\exitwitherr.bat')

    cmd3 = get_cmd('echo three',
                   r'tests\testfiles\cmds\echo.bat three')

    stdout = temp_dir.joinpath('stdout')

    context = Context({
        'cmds': {
            'run': [cmd1, cmd2, cmd3],
            'stdout': stdout,
            'stderr': '/dev/stdout'}
    })

    with pytest.raises(MultiError) as ex:
        pypyr.steps.cmds.run_step(context)

    assert 'cmdOut' not in context
    err = ex.value
    assert len(err) == 1
    the_err = err[0]
    assert type(the_err) is SubprocessError
    assert the_err.cmd == [cmd2]
    assert the_err.returncode == 1
    # None because save is False
    assert the_err.stderr is None

    out_file_lines = stdout.read_text().rstrip().split('\n')
    TestCase().assertCountEqual(out_file_lines,
                                ['one', 'arb err here', 'three'])


def test_async_cmds_overwrite_vs_append(temp_dir):
    """Write to stdout and stderr with append true/false, default overwrite."""
    stdout = temp_dir.joinpath('mydir/stdout')
    stderr = temp_dir.joinpath('mydir/stderr')

    cmd1 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh one',
                   r'tests\testfiles\cmds\echo-out-and-err.bat one')

    context = Context({
        'cmds': {
            'run': [cmd1],
            'stdout': stdout,
            'stderr': stderr}
    })

    pypyr.steps.cmds.run_step(context)

    assert 'cmdOut' not in context
    assert stdout.read_text() == 'stdout one\n'
    assert stderr.read_text() == 'stderr one\n'

    # now overwrite these by default on 2nd invocation
    cmd2 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh two',
                   r'tests\testfiles\cmds\echo-out-and-err.bat two')

    context = Context({
        'cmds': {
            'run': [cmd2],
            'stdout': stdout,
            'stderr': stderr}
    })

    pypyr.steps.cmds.run_step(context)

    assert stdout.read_text() == 'stdout two\n'
    assert stderr.read_text() == 'stderr two\n'

    # now append
    cmd3 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh three',
                   r'tests\testfiles\cmds\echo-out-and-err.bat three')

    # list input with element being a dict in expanded syntax
    context = Context({
        'cmds': [{
            'run': [cmd3],
            'stdout': stdout,
            'stderr': stderr,
            'append': True}]
    })

    pypyr.steps.cmds.run_step(context)

    assert 'cmdOut' not in context
    assert stdout.read_text() == (
        'stdout two\nstdout three\n')
    assert stderr.read_text() == (
        'stderr two\nstderr three\n')


def test_async_cmds_expanded_syntax_on_list_item(temp_dir):
    """Set stdout/stderr on multiple expanded inputs on base list input."""
    cmd1 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh one',
                   r'tests\testfiles\cmds\echo-out-and-err.bat one')

    cmd2 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh two',
                   r'tests\testfiles\cmds\echo-out-and-err.bat two')

    cmd3 = get_cmd('tests/testfiles/cmds/echo-out-and-err.sh three',
                   r'tests\testfiles\cmds\echo-out-and-err.bat three')

    stdout1 = temp_dir.joinpath('stdout1')
    stderr1 = temp_dir.joinpath('stderr1')
    stdout2 = temp_dir.joinpath('stdout2')
    stderr2 = temp_dir.joinpath('stderr2')

    context = Context({
        'cmds': [
            {
                'run': [cmd1],
                'stdout': stdout1,
                'stderr': stderr1
            },
            {
                'run': [cmd2, cmd3],
                'stdout': stdout2,
                'stderr': stderr2}
        ]})

    pypyr.steps.cmds.run_step(context)

    assert 'cmdOut' not in context

    assert stdout1.read_text() == 'stdout one\n'
    assert stderr1.read_text() == 'stderr one\n'

    out_file_lines2 = stdout2.read_text().rstrip().split('\n')
    err_file_lines2 = stderr2.read_text().rstrip().split('\n')

    TestCase().assertCountEqual(out_file_lines2,
                                ['stdout two', 'stdout three'])

    TestCase().assertCountEqual(err_file_lines2,
                                ['stderr two', 'stderr three'])



"""pypyr step that loads toml file into context."""
import logging

import pypyr.toml as toml
from pypyr.utils.asserts import assert_key_has_value

logger = logging.getLogger(__name__)


def run_step(context):
    """Load a toml file into the pypyr context.

    toml parsed from the file will be merged into the pypyr context. This will
    overwrite existing values if the same keys are already in there.
    I.e if file toml has
    eggs = "boiled"
    and context {'eggs': 'fried'} already exists, returned context['eggs'] will
    be 'boiled'.

    Args:
        context: pypyr.context.Context. Mandatory.
                 The following context key must exist
                - fetchToml
                    - path. path-like. Path to file on disk.
                    - key. string. If exists, write toml structure to this
                      context key. Else toml writes to context root.

    Also supports passing a path as string to fetchToml, but in this case you
    won't be able to specify a key.

    Note that TOML must always be utf-8, per spec.

    All inputs support formatting expressions.

    Returns:
        None. updates context arg.

    Raises:
        FileNotFoundError: take a guess
        pypyr.errors.KeyNotInContextError: fetchToml.path missing in context.
        pypyr.errors.KeyInContextHasNoValueError: fetchToml.path exists but is
                                                  None.

    """
    logger.debug("started")

    context.assert_key_has_value(key='fetchToml', caller=__name__)

    fetch_toml_input = context.get_formatted('fetchToml')

    if isinstance(fetch_toml_input, str):
        file_path = fetch_toml_input
        destination_key = None
    else:
        assert_key_has_value(obj=fetch_toml_input,
                             key='path',
                             caller=__name__,
                             parent='fetchToml')
        file_path = fetch_toml_input['path']
        destination_key = fetch_toml_input.get('key', None)

    logger.debug("attempting to open file: %s", file_path)

    payload = toml.read_file(file_path)

    if destination_key:
        logger.debug("toml file loaded. Writing to context %s",
                     destination_key)
        context[destination_key] = payload
    else:
        # toml by definition has Mapping on top level, no need to check,
        logger.debug("toml file loaded. Merging into pypyr context...")
        context.update(payload)

    logger.info("toml file written into pypyr context. Count: %s",
                len(payload))
    logger.debug("done")



"""tomlfile.py unit tests."""
from unittest.mock import mock_open, patch

import pytest

import pypyr.parser.tomlfile as toml_file


def test_toml_parser_file_open_fails_on_arbitrary_string():
    """Non path-y input string should fail."""
    with pytest.raises(FileNotFoundError):
        toml_file.get_parsed_context('value 1,value 2, value3')


def test_toml__parser_file_open_returns_none_on_empty_string():
    """Non path-y input string should fail."""
    assert toml_file.get_parsed_context('') is None


def test_toml_parser_file_open_returns_none_on_none():
    """Non path-y input string should fail."""
    assert toml_file.get_parsed_context(None) is None


def test_toml_parser_pass():
    """Path should pass to toml parser and return parsed toml as dict."""
    in_bytes = b'[table]\nkey= "value"'

    with patch('pypyr.toml.open',
               mock_open(read_data=in_bytes)) as mocked_open:
        out = toml_file.get_parsed_context(['./myfile.toml'])

    mocked_open.assert_called_once_with('./myfile.toml', 'rb')

    assert out == {'table': {'key': 'value'}}


def test_toml_parser_spaces():
    """Unescaped path with spaces pass."""
    in_bytes = b'key= "value"'

    with patch('pypyr.toml.open',
               mock_open(read_data=in_bytes)) as mocked_open:
        out = toml_file.get_parsed_context(['/mydir/sub/one',
                                            'two/my',
                                            'file.toml'])

    mocked_open.assert_called_once_with('/mydir/sub/one two/my file.toml',
                                        'rb')

    assert out == {'key': 'value'}



"""pypyr.loaders.string integration tests."""
from textwrap import dedent

from pypyr import pipelinerunner
from pypyr.loaders.string import get_pipeline_definition


def test_get_pipeline_definition():
    """Test get_pipeline_definition."""
    pipeline = dedent(
        """\
        steps:
          - name: pypyr.steps.echo
            in:
              echoMe: 'testing'
        """
    )

    definition = get_pipeline_definition(pipeline, None)

    assert definition.pipeline == {
        "steps": [{"name": "pypyr.steps.echo", "in": {"echoMe": "testing"}}]
    }


def test_run_with_custom_runner():
    """Test run with custom string loader."""
    pipeline = dedent(
        """\
        steps:
          - name: pypyr.steps.set
            in:
              set:
                test: 1
        """
    )

    context = pipelinerunner.run(
        pipeline_name=pipeline, loader="pypyr.loaders.string"
    )

    assert context["test"] == 1



"""nowutc.py unit tests."""
from datetime import datetime, timezone
from unittest.mock import patch, Mock
from pypyr.context import Context
import pypyr.steps.nowutc as nowutc_step

frozen_timestamp = datetime.now(timezone.utc)


@patch('pypyr.steps.nowutc.datetime',
       Mock(now=Mock(return_value=frozen_timestamp)))
def test_nowutc_default_iso():
    """Now gets date in iso format."""
    context = Context()
    nowutc_step.run_step(context)

    assert context['nowUtc'] == frozen_timestamp.isoformat()


@patch('pypyr.steps.nowutc.datetime',
       Mock(now=Mock(return_value=frozen_timestamp)))
def test_nowutc_with_formatting_date_part():
    """Now gets timestamp with formatting."""
    context = Context({'nowUtcIn': '%Y%m%d'})
    nowutc_step.run_step(context)

    assert context['nowUtc'] == frozen_timestamp.strftime('%Y%m%d')


@patch('pypyr.steps.nowutc.datetime',
       Mock(now=Mock(return_value=frozen_timestamp)))
def test_nowutc_with_formatting_time_part():
    """Now gets timestamp with date formatting."""
    context = Context({'nowUtcIn': '%I hour %M %p blah %Z'})
    nowutc_step.run_step(context)

    assert context['nowUtc'] == frozen_timestamp.strftime(
        '%I hour %M %p blah UTC')


@patch('pypyr.steps.nowutc.datetime',
       Mock(now=Mock(return_value=frozen_timestamp)))
def test_nowutc_with_formatting_interpolation():
    """Now gets timestamp with date formatting and pypyr interpolation."""
    context = Context({'f': '%Y %w', 'nowUtcIn': '%A {f}'})
    nowutc_step.run_step(context)

    assert context['nowUtc'] == frozen_timestamp.strftime('%A %Y %w')



"""Unit tests for pypyr.steps.dsl.venv."""
from types import SimpleNamespace
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch, call, DEFAULT

import pytest

from pypyr.context import Context
from pypyr.errors import ContextError, KeyNotInContextError, MultiError
from pypyr.steps.dsl.venv import (
    EXCEPTION_INPUT,
    EXCEPTION_LIST_INPUT,
    VenvCreatorStep)


def get_simple_context():
    """Create a test SimpleNamespace to serve as EnvBuilder context."""
    context = SimpleNamespace()
    context.env_exec_cmd = '/python'
    context.bin_path = '/venv/bin'
    context.env_dir = '/venv'
    return context


def test_venv_dsl_no_config_key():
    """Raise error when required venv key missing."""
    with pytest.raises(KeyNotInContextError) as err:
        VenvCreatorStep.from_context(Context({'a': 'b'}))

    assert str(err.value) == (
        "context['venv'] doesn't exist. "
        + "It must exist for pypyr.steps.venv.")


def test_venv_dsl_context_wrong_type():
    """Raise error when required venv is wrong type."""
    with pytest.raises(ContextError) as err:
        VenvCreatorStep.from_context(Context({'venv': 123}))

    assert str(err.value) == EXCEPTION_INPUT.format(bad_type='int')


def test_venv_dsl_context_list_contains_wrong_type():
    """Raise error when required venv is list with a wrong type."""
    with pytest.raises(ContextError) as err:
        VenvCreatorStep.from_context(Context({'venv': ['/arb', 123]}))

    assert str(err.value) == EXCEPTION_LIST_INPUT.format(bad_type='int')


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_simple_str(mock_builder):
    """Run with simple str input."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context

    step = VenvCreatorStep.from_context(Context({'venv': '/arb'}))

    expected_path = str(Path('/arb').expanduser().resolve())
    assert len(step.venvs) == 1
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path

    step.run_step()

    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_called_once_with(context)
    mocked_builder.pip_install_extras.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_mapping_simple_str(mock_builder):
    """Run with simple str input inside mapping."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context

    step = VenvCreatorStep.from_context(Context(
        {'venv': {
            'path': '/arb'}}))

    expected_path = str(Path('/arb').expanduser().resolve())
    assert len(step.venvs) == 1
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path

    step.run_step()

    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_called_once_with(context)
    mocked_builder.pip_install_extras.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_mapping_list_of_str(mock_builder):
    """Run with list of str input inside mapping."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context

    step = VenvCreatorStep.from_context(Context(
        {'venv': {
            'path': ['/arb1', '/arb2']}}))

    expected_path1 = str(Path('/arb1').expanduser().resolve())
    expected_path2 = str(Path('/arb2').expanduser().resolve())

    assert len(step.venvs) == 2
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path1

    venv_creator = step.venvs[1]
    assert venv_creator.path == expected_path2

    assert mock_builder.call_count == 2

    step.run_step()

    assert mocked_builder.create.call_count == 2
    assert sorted(mocked_builder.create.mock_calls) == sorted(
        [call(expected_path1),
         call(expected_path2)])

    assert mocked_builder.upgrade_dependencies.call_count == 2
    assert mocked_builder.upgrade_dependencies.mock_calls[0] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[1] == call(context)

    mocked_builder.pip_install_extras.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_list_of_str(mock_builder):
    """Run with list of str input."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context

    step = VenvCreatorStep.from_context(Context(
        {'venv': ['/arb1', '/arb2']}))

    expected_path1 = str(Path('/arb1').expanduser().resolve())
    expected_path2 = str(Path('/arb2').expanduser().resolve())

    assert len(step.venvs) == 2
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path1

    venv_creator = step.venvs[1]
    assert venv_creator.path == expected_path2

    assert mock_builder.call_count == 2

    step.run_step()

    assert mocked_builder.create.call_count == 2
    assert sorted(mocked_builder.create.mock_calls) == sorted(
        [call(expected_path1),
         call(expected_path2)])

    assert mocked_builder.upgrade_dependencies.call_count == 2
    assert mocked_builder.upgrade_dependencies.mock_calls[0] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[1] == call(context)

    mocked_builder.pip_install_extras.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_list_of_str_and_mapping(mock_builder):
    """Run with list of str and mapping mixed input."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context

    step = VenvCreatorStep.from_context(Context(
        {'venv': ['/arb1',
                  {'path': '/arb2'}]}))

    expected_path1 = str(Path('/arb1').expanduser().resolve())
    expected_path2 = str(Path('/arb2').expanduser().resolve())

    assert len(step.venvs) == 2
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path1

    venv_creator = step.venvs[1]
    assert venv_creator.path == expected_path2

    assert mock_builder.call_count == 2

    step.run_step()

    assert mocked_builder.create.call_count == 2
    assert sorted(mocked_builder.create.mock_calls) == sorted([
        call(expected_path1),
        call(expected_path2)])

    assert mocked_builder.upgrade_dependencies.call_count == 2
    assert mocked_builder.upgrade_dependencies.mock_calls[0] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[1] == call(context)

    mocked_builder.pip_install_extras.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_list_of_str_and_mapping_list(mock_builder):
    """Run with list of str and mapping containing list path mixed input."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context

    step = VenvCreatorStep.from_context(Context(
        {'venv': ['/arb1',
                  {'path': ['/arb2', '/arb3']}]}))

    expected_path1 = str(Path('/arb1').expanduser().resolve())
    expected_path2 = str(Path('/arb2').expanduser().resolve())
    expected_path3 = str(Path('/arb3').expanduser().resolve())

    assert len(step.venvs) == 3
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path1

    venv_creator = step.venvs[1]
    assert venv_creator.path == expected_path2

    venv_creator = step.venvs[2]
    assert venv_creator.path == expected_path3

    assert mock_builder.call_count == 3

    step.run_step()

    assert mocked_builder.create.call_count == 3
    assert sorted(mocked_builder.create.mock_calls) == sorted(
        [call(expected_path1),
         call(expected_path2),
         call(expected_path3)])

    assert mocked_builder.upgrade_dependencies.call_count == 3
    assert mocked_builder.upgrade_dependencies.mock_calls[0] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[1] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[2] == call(context)

    mocked_builder.pip_install_extras.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_list_of_str_and_mapping_list_no_pip(mock_builder):
    """Run with with_pip False does not install dependencies."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context

    step = VenvCreatorStep.from_context(Context(
        {'venv': ['/arb1',
                  {'path': ['/arb2', '/arb3'],
                   'with_pip': False}]}))

    expected_path1 = str(Path('/arb1').expanduser().resolve())
    expected_path2 = str(Path('/arb2').expanduser().resolve())
    expected_path3 = str(Path('/arb3').expanduser().resolve())

    assert len(step.venvs) == 3
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path1

    venv_creator = step.venvs[1]
    assert venv_creator.path == expected_path2

    venv_creator = step.venvs[2]
    assert venv_creator.path == expected_path3

    assert mock_builder.call_count == 3

    step.run_step()

    assert mocked_builder.create.call_count == 3
    assert sorted(mocked_builder.create.mock_calls) == sorted(
        [call(expected_path1),
         call(expected_path2),
         call(expected_path3)])

    mocked_builder.upgrade_dependencies.assert_called_once_with(context)

    mocked_builder.pip_install_extras.assert_not_called()


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_simple_str_error(mock_builder):
    """Run with simple str input raising error on create."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context
    mocked_builder.create.side_effect = ValueError('arb')

    step = VenvCreatorStep.from_context(Context({'venv': '/arb'}))

    expected_path = str(Path('/arb').expanduser().resolve())

    assert len(step.venvs) == 1
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path

    with pytest.raises(ValueError) as err:
        step.run_step()

    assert str(err.value) == 'arb'

    mocked_builder.create.assert_called_once_with(expected_path)
    mocked_builder.upgrade_dependencies.assert_not_called()
    mocked_builder.pip_install_extras.assert_not_called()


arb2_path = str(Path('/arb2').expanduser().resolve())
arb4_path = str(Path('/arb4').expanduser().resolve())


def env_builder_create_mock(path):
    """Force the correct order even for async calls."""
    if path == arb2_path:
        raise ValueError('arb')
    if path == arb4_path:
        raise ValueError('arb2')


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_mapping_list_of_str_error_on_create(mock_builder):
    """Run with list of str input inside mapping raising create errors."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context
    mocked_builder.create.side_effect = env_builder_create_mock

    step = VenvCreatorStep.from_context(Context(
        {'venv': {
            'path': ['/arb1', '/arb2', '/arb3', '/arb4'],
            'pip': 'package1 package2'}}))

    expected_path1 = str(Path('/arb1').expanduser().resolve())
    expected_path2 = str(Path('/arb2').expanduser().resolve())
    expected_path3 = str(Path('/arb3').expanduser().resolve())
    expected_path4 = str(Path('/arb4').expanduser().resolve())

    assert len(step.venvs) == 4
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path1

    venv_creator = step.venvs[1]
    assert venv_creator.path == expected_path2

    venv_creator = step.venvs[2]
    assert venv_creator.path == expected_path3

    venv_creator = step.venvs[3]
    assert venv_creator.path == expected_path4

    assert mock_builder.call_count == 4

    with pytest.raises(MultiError) as err:
        step.run_step()

    assert len(err.value.errors) == 2
    assert repr(err.value.errors[0]) == repr(ValueError('arb'))
    assert repr(err.value.errors[1]) == repr(ValueError('arb2'))

    assert mocked_builder.create.call_count == 4
    assert sorted(mocked_builder.create.mock_calls) == sorted(
        [call(expected_path1),
         call(expected_path2),
         call(expected_path3),
         call(expected_path4)])

    assert mocked_builder.upgrade_dependencies.call_count == 2
    assert mocked_builder.upgrade_dependencies.mock_calls[0] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[1] == call(context)

    assert mocked_builder.pip_install_extras.call_count == 2
    assert mocked_builder.pip_install_extras.mock_calls[0] == call(
        'package1 package2')
    assert mocked_builder.pip_install_extras.mock_calls[1] == call(
        'package1 package2')


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_list_of_mapping_error_on_create(mock_builder):
    """Run with input list of mapping varying pip raising create errors."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context
    mocked_builder.create.side_effect = env_builder_create_mock

    step = VenvCreatorStep.from_context(Context(
        {'venv': [
            {'path': '/arb1',
             'pip': 'package1 package2'},
            {'path': '/arb2',
             'pip': 'package3 package4'},
            {'path': '/arb3',
             'pip': 'package5 package6'},
            {'path': '/arb4',
             'pip': 'package7 package8'},
        ]}))

    expected_path1 = str(Path('/arb1').expanduser().resolve())
    expected_path2 = str(Path('/arb2').expanduser().resolve())
    expected_path3 = str(Path('/arb3').expanduser().resolve())
    expected_path4 = str(Path('/arb4').expanduser().resolve())

    assert len(step.venvs) == 4
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path1

    venv_creator = step.venvs[1]
    assert venv_creator.path == expected_path2

    venv_creator = step.venvs[2]
    assert venv_creator.path == expected_path3

    venv_creator = step.venvs[3]
    assert venv_creator.path == expected_path4

    assert mock_builder.call_count == 4

    with pytest.raises(MultiError) as err:
        step.run_step()

    assert len(err.value.errors) == 2
    assert repr(err.value.errors[0]) == repr(ValueError('arb'))
    assert repr(err.value.errors[1]) == repr(ValueError('arb2'))

    assert mocked_builder.create.call_count == 4
    assert sorted(mocked_builder.create.mock_calls) == sorted(
        [call(expected_path1),
         call(expected_path2),
         call(expected_path3),
         call(expected_path4)])

    assert mocked_builder.upgrade_dependencies.call_count == 2
    assert mocked_builder.upgrade_dependencies.mock_calls[0] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[1] == call(context)

    assert mocked_builder.pip_install_extras.call_count == 2
    assert mocked_builder.pip_install_extras.mock_calls[0] == call(
        'package1 package2')
    assert mocked_builder.pip_install_extras.mock_calls[1] == call(
        'package5 package6')


@patch('pypyr.venv.EnvBuilderWithExtraDeps')
def test_venv_dsl_mapping_list_of_str_error_on_pip_install(mock_builder):
    """Run with list of str input inside mapping raising extras pip errors."""
    context = get_simple_context()
    mocked_builder = mock_builder.return_value
    mocked_builder.context = context
    mocked_builder.pip_install_extras.side_effect = [DEFAULT,
                                                     ValueError('arb'),
                                                     DEFAULT,
                                                     ValueError('arb2')]

    step = VenvCreatorStep.from_context(Context(
        {'venv': [
            {'path': '/arb1',
             'pip': 'package1 package2'},
            {'path': '/arb2',
             'pip': 'package3 package4'},
            {'path': '/arb3',
             'pip': 'package5 package6'},
            {'path': '/arb4',
             'pip': 'package7 package8'},
        ]}))

    expected_path1 = str(Path('/arb1').expanduser().resolve())
    expected_path2 = str(Path('/arb2').expanduser().resolve())
    expected_path3 = str(Path('/arb3').expanduser().resolve())
    expected_path4 = str(Path('/arb4').expanduser().resolve())

    assert len(step.venvs) == 4
    venv_creator = step.venvs[0]
    assert venv_creator.path == expected_path1

    venv_creator = step.venvs[1]
    assert venv_creator.path == expected_path2

    venv_creator = step.venvs[2]
    assert venv_creator.path == expected_path3

    venv_creator = step.venvs[3]
    assert venv_creator.path == expected_path4

    assert mock_builder.call_count == 4

    with pytest.raises(MultiError) as err:
        step.run_step()

    assert len(err.value.errors) == 2
    assert repr(err.value.errors[0]) == repr(ValueError('arb'))
    assert repr(err.value.errors[1]) == repr(ValueError('arb2'))

    assert mocked_builder.create.call_count == 4
    assert sorted(mocked_builder.create.mock_calls) == sorted(
        [call(expected_path1),
         call(expected_path2),
         call(expected_path3),
         call(expected_path4)])

    assert mocked_builder.upgrade_dependencies.call_count == 4
    assert mocked_builder.upgrade_dependencies.mock_calls[0] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[1] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[2] == call(context)
    assert mocked_builder.upgrade_dependencies.mock_calls[3] == call(context)

    assert mocked_builder.pip_install_extras.call_count == 4
    # strictly speaking the mock_calls list could be in any order, because
    # envs might have finished creating out of instantiation order.
    TestCase().assertCountEqual(
        mocked_builder.pip_install_extras.mock_calls,
        ([
            call('package1 package2'),
            call('package3 package4'),
            call('package5 package6'),
            call('package7 package8')]))
